/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid.attacheditem;

import static java.util.Objects.*;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.collections.MapUtils;
import org.springframework.stereotype.Service;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.command.Command;
import de.ozgcloud.common.errorhandling.TechnicalException;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service
@RequiredArgsConstructor
@Log4j2
public class AttachedItemService {

	public static final String BESCHEID_ITEM_NAME = "Bescheid";
	static final String CREATE_ATTACHED_ITEM_ORDER = "CREATE_ATTACHED_ITEM";
	static final String UPDATE_ATTACHED_ITEM_ORDER = "UPDATE_ATTACHED_ITEM";
	static final String DELETE_ATTACHED_ITEM = "DELETE_ATTACHED_ITEM";

	private static final Predicate<String> notExpectedSendByValue = sendBy -> !BescheidItem.ACCEPTED_SEND_BY_VALUES.contains(sendBy);

	private final OzgCloudCommandService commandService;
	private final VorgangAttachedItemRemoteService remoteService;
	private final CommandMapper commandMapper;

	public Optional<BescheidItem> findBescheidItem(String vorgangId) {
		return remoteService.findBescheidDraft(vorgangId);
	}

	public BescheidItem getBescheidItem(String id) {
		return remoteService.getBescheid(id);
	}

	public String createBescheidDraft(Command command) {
		validateBescheidData(command.getBodyObject());
		return remoteService.findBescheidDraft(command.getVorgangId())
				.map(bescheid -> overrideAttachedItem(bescheid, command)).orElseGet(() -> createAttachedItem(command));
	}

	String overrideAttachedItem(BescheidItem bescheidItem, Command command) {
		var bodyObject = buildAttachedItemAsMap(command, buildItemMapWithAllBescheidFields(command));
		var finishedOzgCloudCommand = commandService.createAndWaitUntilDone(buildUpdateAttachedItemCommand(bescheidItem, bodyObject));
		return finishedOzgCloudCommand.getCreatedResource();
	}

	Map<String, Object> buildItemMapWithAllBescheidFields(Command command) {
		var result = new HashMap<String, Object>();
		result.put(BescheidItem.FIELD_BESCHIEDEN_AM, command.getBodyObject().get(BescheidItem.FIELD_BESCHIEDEN_AM));
		result.put(BescheidItem.FIELD_BEWILLIGT, command.getBodyObject().get(BescheidItem.FIELD_BEWILLIGT));
		result.put(BescheidItem.FIELD_BESCHEID_DOCUMENT, command.getBodyObject().get(BescheidItem.FIELD_BESCHEID_DOCUMENT));
		result.put(BescheidItem.FIELD_SEND_BY, command.getBodyObject().get(BescheidItem.FIELD_SEND_BY));
		result.put(BescheidItem.FIELD_NACHRICHT_TEXT, command.getBodyObject().get(BescheidItem.FIELD_NACHRICHT_TEXT));
		result.put(BescheidItem.FIELD_NACHRICHT_SUBJECT, command.getBodyObject().get(BescheidItem.FIELD_NACHRICHT_SUBJECT));
		result.put(BescheidItem.FIELD_ATTACHMENTS, command.getBodyObject().get(BescheidItem.FIELD_ATTACHMENTS));
		return result;
	}

	public void updateBescheidDraft(Command command) {
		var bescheidItem = remoteService.getBescheid(command.getRelationId());
		validateBescheidStatus(bescheidItem);
		var itemMap = buildItemMapWithExistingBescheidFields(command);
		var bodyObject = buildAttachedItemAsMap(command, itemMap);
		commandService.createAndWaitUntilDone(buildUpdateAttachedItemCommand(bescheidItem, bodyObject));
	}

	OzgCloudCommand buildUpdateAttachedItemCommand(BescheidItem bescheidItem, Map<String, Object> bodyObject) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(bescheidItem.getVorgangId()))
				.relationId(commandMapper.mapRelationId(bescheidItem.getId()))
				.relationVersion(bescheidItem.getVersion())
				.order(UPDATE_ATTACHED_ITEM_ORDER)
				.bodyObject(bodyObject)
				.build();
	}

	String createAttachedItem(Command command) {
		var finishedOzgCloudCommand = commandService.createAndWaitUntilDone(buildCreateAttachedItemCommand(command));
		return finishedOzgCloudCommand.getCreatedResource();
	}

	OzgCloudCommand buildCreateAttachedItemCommand(Command command) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(command.getVorgangId()))
				.relationId(commandMapper.mapRelationId(command.getRelationId()))
				.order(CREATE_ATTACHED_ITEM_ORDER)
				.bodyObject(buildAttachedItemAsMap(command, buildItemMapWithExistingBescheidFields(command))).build();
	}

	Map<String, Object> buildAttachedItemAsMap(Command command, Map<String, Object> itemmap) {
		var result = new HashMap<String, Object>();
		result.put(BescheidItem.PROPERTY_VORGANG_ID, command.getVorgangId());
		result.put(BescheidItem.PROPERTY_CLIENT, BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		result.put(BescheidItem.PROPERTY_ITEM_NAME, BESCHEID_ITEM_NAME);
		result.put(BescheidItem.PROPERTY_ITEM, itemmap);
		return result;
	}

	Map<String, Object> buildItemMapWithExistingBescheidFields(Command command) {
		var result = new HashMap<String, Object>();
		result.put(BescheidItem.FIELD_STATUS, BescheidItem.Status.DRAFT.name());
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_BESCHIEDEN_AM, result);
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_BEWILLIGT, result);
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_BESCHEID_DOCUMENT, result);
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_SEND_BY, result);
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_NACHRICHT_TEXT, result);
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_NACHRICHT_SUBJECT, result);
		addValueFromMapIfExists(command.getBodyObject(), BescheidItem.FIELD_ATTACHMENTS, result);
		return result;
	}

	void addValueFromMapIfExists(Map<String, Object> sourceMap, String key, Map<String, Object> targetMap) {
		if (sourceMap.containsKey(key)) {
			targetMap.put(key, sourceMap.get(key));
		}
	}

	void validateBescheidData(Map<String, Object> bodyObject) {
		if (isNull(bodyObject.get(BescheidItem.FIELD_BESCHIEDEN_AM))) {
			throw new TechnicalException("Fields '%s' is required for bescheid creation".formatted(BescheidItem.FIELD_BESCHIEDEN_AM));
		}
		if (isNull(bodyObject.get(BescheidItem.FIELD_BEWILLIGT))) {
			throw new TechnicalException("Fields '%s' is required for bescheid creation".formatted(BescheidItem.FIELD_BEWILLIGT));
		}
		Optional.ofNullable(MapUtils.getString(bodyObject, BescheidItem.FIELD_SEND_BY)).filter(notExpectedSendByValue)
				.ifPresent(sendBy ->
						LOG.warn("Unexpected value for field '%s': %s. Allowed are: %s".formatted(BescheidItem.FIELD_SEND_BY, sendBy,
								String.join(", ", BescheidItem.ACCEPTED_SEND_BY_VALUES)))
				);
	}

	public void deleteBescheidDraft(Command command) {
		var bescheid = remoteService.getBescheid(command.getRelationId());
		validateBescheidStatus(bescheid);
		commandService.createAndWaitUntilDone(buildDeleteItemCommand(command, bescheid));
	}

	void validateBescheidStatus(BescheidItem bescheid) {
		var bescheidStatus = MapUtils.getString(bescheid.getBescheidData(), BescheidItem.FIELD_STATUS);
		if (BescheidItem.Status.DRAFT.not(bescheidStatus)) {
			throw new TechnicalException("Bescheid draft with ID '%s' has an unexpected status: '%s'".formatted(bescheid.getId(), bescheidStatus));
		}
	}

	OzgCloudCommand buildDeleteItemCommand(Command command, BescheidItem bescheid) {
		return OzgCloudCommand.builder()
				.vorgangId(commandMapper.toOzgCloudVorgangId(command.getVorgangId()))
				.relationId(commandMapper.mapRelationId(bescheid.getId()))
				.relationVersion(bescheid.getVersion())
				.order(DELETE_ATTACHED_ITEM)
				.build();
	}

	public AttachedItem getItem(String id) {
		return remoteService.getItem(id);
	}
}
