package de.ozgcloud.bescheid.nachricht;

import java.io.IOException;
import java.io.StringWriter;
import java.util.Objects;
import java.util.Optional;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.bescheid.Bescheid;
import de.ozgcloud.bescheid.vorgang.Vorgang.PostfachAddress;
import de.ozgcloud.common.errorhandling.TechnicalException;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
public class NachrichtService {

	@Autowired
	private NachrichtRemoteService remoteService;

	@Autowired
	private Configuration freemarkerCfg;

	static final String SUBJECT = "Ihr Antrag";

	private static final String TEMPLATE_FILE = "bescheid.nachrichtTemplate.txt.ftlh";

	public void createNachrichtDraft(Bescheid bescheid) {
		buildNachricht(bescheid).ifPresentOrElse(remoteService::saveDraft, () -> LOG.warn("No ServiceKonto given on Vorgang."));
	}

	Optional<Nachricht> buildNachricht(Bescheid bescheid) {
		return getAddress(bescheid).map(address -> Nachricht.builder()
				.vorgangId(bescheid.getVorgangId())
				.postfachAddress(address)
				.subject(SUBJECT)
				.mailBody(buildMessage(bescheid))
				.createdBy(bescheid.getCreatedBy())
				.bescheidFileId(bescheid.getBescheidFileId())
				.build());
	}

	Optional<PostfachAddress> getAddress(Bescheid bescheid) {
		var serviceKonto = bescheid.getServiceKonto();
		if (Objects.nonNull(serviceKonto)) {
			return Optional.of(serviceKonto.getPostfachAddresses().get(0));
		}
		return Optional.empty();
	}

	String buildMessage(Bescheid bescheid) {
		return bescheid.getNachrichtText()
				.orElseGet(() -> fillTemplate(TEMPLATE_FILE, bescheid));
	}

	String fillTemplate(String templateName, Object dataModel) {
		try {
			Template template = getTemplate(templateName);
			StringWriter stringWriter = new StringWriter();
			template.process(dataModel, stringWriter);
			return stringWriter.toString();

		} catch (IOException | TemplateException e) {
			throw new TechnicalException("Error filling template", e);
		}
	}

	Template getTemplate(String templateName) {
		try {
			return freemarkerCfg.getTemplate(templateName);
		} catch (IOException e) {
			throw new TechnicalException("Error loading mail template", e);
		}
	}
}
