/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid.attacheditem;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.apilib.common.command.OzgCloudCommand;
import de.ozgcloud.apilib.common.command.OzgCloudCommandService;
import de.ozgcloud.apilib.common.command.grpc.CommandMapper;
import de.ozgcloud.apilib.common.datatypes.GenericId;
import de.ozgcloud.apilib.vorgang.OzgCloudVorgangId;
import de.ozgcloud.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import io.grpc.StatusRuntimeException;

class AttachedItemServiceTest {

	@Spy
	@InjectMocks
	private AttachedItemService service;

	@Mock
	private OzgCloudCommandService commandService;
	@Mock
	private CommandMapper commandMapper;
	@Mock
	private VorgangAttachedItemRemoteService remoteService;

	@Nested
	class TestFindBescheidItem {

		@Test
		void shouldCallGetBescheidDraft() {
			service.findBescheidItem(CommandTestFactory.VORGANG_ID);

			verify(remoteService).findBescheidDraft(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldReturnBescheidItem() {
			var expected = BescheidItemTestFactory.create();
			when(remoteService.findBescheidDraft(anyString())).thenReturn(Optional.of(expected));

			var actual = service.findBescheidItem(CommandTestFactory.VORGANG_ID);

			assertThat(actual).contains(expected);
		}
	}

	@Nested
	class TestGetBescheidItem {

		@Test
		void shouldCallGetBescheid() {
			service.getBescheidItem(BescheidItemTestFactory.ID);

			verify(remoteService).getBescheid(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldReturnBescheidItem() {
			var expected = BescheidItemTestFactory.create();
			when(remoteService.getBescheid(anyString())).thenReturn(expected);

			var actual = service.getBescheidItem(BescheidItemTestFactory.ID);

			assertThat(actual).isEqualTo(expected);
		}
	}

	@Nested
	class TestCreateBescheidDraft {

		private static final String CREATED_ATTACHED_ITEM_ID = "attached-item-id";

		private Command command = CommandTestFactory.createBuilder().bodyObject(BescheidItemTestFactory.createBescheidBody()).build();
		private BescheidItem bescheidItem = BescheidItemTestFactory.create();

		@Mock
		private OzgCloudCommand ozgCloudCommand;

		@Captor
		private ArgumentCaptor<BescheidItem> bescheidItemCaptor;

		@Test
		void shouldCallValidateData() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createAttachedItem(any());

			service.createBescheidDraft(command);

			verify(service).validateBescheidData(command.getBodyObject());
		}

		@Test
		void shouldCallGetBescheidDraft() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createAttachedItem(any());

			service.createBescheidDraft(command);

			verify(remoteService).findBescheidDraft(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldCallUpdateBescheidDraft() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).overrideAttachedItem(any(), any());
			when(remoteService.findBescheidDraft(anyString())).thenReturn(Optional.of(bescheidItem));

			service.createBescheidDraft(command);

			verify(service).overrideAttachedItem(bescheidItemCaptor.capture(), eq(command));
			assertThat(bescheidItemCaptor.getValue()).isEqualTo(bescheidItem);
		}

		@Test
		void shouldCallCreateBescheidDraft() {
			doReturn(CREATED_ATTACHED_ITEM_ID).when(service).createAttachedItem(any());

			service.createBescheidDraft(command);

			verify(service).createBescheidDraft(command);
		}

		@Nested
		class TestCreateAttachedItem {

			@Mock
			private OzgCloudCommand createAttachedItemCommand;

			@BeforeEach
			void init() {
				when(commandService.createAndWaitUntilDone(any())).thenReturn(ozgCloudCommand);
			}

			@Test
			void shouldCallBuildCreateAttachedItemCommand() {
				createAttachedItem();

				verify(service).buildCreateAttachedItemCommand(command);
			}

			@Test
			void shouldCallCommandService() {
				doReturn(createAttachedItemCommand).when(service).buildCreateAttachedItemCommand(any());

				createAttachedItem();

				verify(commandService).createAndWaitUntilDone(createAttachedItemCommand);
			}

			private void createAttachedItem() {
				service.createAttachedItem(command);
			}
		}

		@Nested
		class TestBuildCreateAttachedItemCommand {

			@Test
			void shouldCallVorgangIdMapper() {
				service.buildCreateAttachedItemCommand(command);

				verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
			}

			@Test
			void shouldSetVorgangId() {
				var expectedVorgangId = OzgCloudVorgangId.from(CommandTestFactory.VORGANG_ID);
				when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

				var result = service.buildCreateAttachedItemCommand(command);

				assertThat(result.getVorgangId()).isEqualTo(expectedVorgangId);
			}

			@Test
			void shouldCallRelationIdMapper() {
				var expectedRelationId = "relationId";
				command = CommandTestFactory.createBuilder().relationId(expectedRelationId).build();
				when(commandMapper.mapRelationId(any())).thenReturn(GenericId.from(command.getRelationId()));

				service.buildCreateAttachedItemCommand(command);

				verify(commandMapper).mapRelationId(expectedRelationId);
			}

			@Test
			void shouldSetRelationId() {
				var expectedRelationId = "relationId";
				command = CommandTestFactory.createBuilder().relationId(expectedRelationId).build();
				when(commandMapper.mapRelationId(any())).thenReturn(GenericId.from(command.getRelationId()));

				var result = service.buildCreateAttachedItemCommand(command);

				assertThat(result.getRelationId()).hasToString(command.getRelationId());
			}

			@Test
			void shouldCallBuildAttachedItemAsMap() {
				var expectedItemMap = Map.of("key", (Object) "value");
				doReturn(expectedItemMap).when(service).buildItemMapWithExistingBescheidFields(any());

				service.buildCreateAttachedItemCommand(command);

				verify(service).buildAttachedItemAsMap(command, expectedItemMap);
			}

			@Test
			void shouldSetBodyObject() {
				var expectedBodyObject = Map.of("key", (Object) "value");
				doReturn(expectedBodyObject).when(service).buildAttachedItemAsMap(any(), any());

				var result = service.buildCreateAttachedItemCommand(command);

				assertThat(result.getBodyObject()).containsAllEntriesOf(expectedBodyObject);
			}

			@Test
			void shouldSetOrder() {
				var result = service.buildCreateAttachedItemCommand(command);

				assertThat(result.getOrder()).isEqualTo(AttachedItemService.CREATE_ATTACHED_ITEM_ORDER);
			}
		}

		@Nested
		class TestValidateBescheidData {

			@Test
			void shouldValidate() {
				assertDoesNotThrow(() -> service.validateBescheidData(command.getBodyObject()));
			}

			@DisplayName(value = "should fail if bescheidAm is missing")
			@Test
			void shouldThrowExceptionByBescheidAm() {
				Map<String, Object> map = Map.of(BescheidItem.FIELD_BEWILLIGT, true);

				assertThrows(TechnicalException.class, () -> service.validateBescheidData(map));
			}

			@DisplayName(value = "should fail if bewilligt is missing")
			@Test
			void shouldThrowExceptionByBewilligt() {
				Map<String, Object> map = Map.of(BescheidItem.FIELD_BESCHIEDEN_AM, "2021-01-01");

				assertThrows(TechnicalException.class, () -> service.validateBescheidData(map));
			}
		}
	}

	@Nested
	class TestUpdateBescheidDraft {

		@Mock
		private OzgCloudCommand updateAttachedItemCommand;

		private Command command = CommandTestFactory.createBuilder().relationId(BescheidItemTestFactory.ID).build();
		private BescheidItem bescheidItem = BescheidItemTestFactory.create();

		@BeforeEach
		void init() {
			when(remoteService.getBescheid(any())).thenReturn(bescheidItem);
		}

		@Test
		void shouldCallGetBescheid() {
			service.updateBescheidDraft(command);

			verify(remoteService).getBescheid(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldCallValidateBescheidStatus() {
			service.updateBescheidDraft(command);

			verify(service).validateBescheidStatus(bescheidItem);
		}

		@Test
		void shouldCallBuildItemMapWithExistingBescheidFields() {
			service.updateBescheidDraft(command);

			verify(service).buildItemMapWithExistingBescheidFields(command);
		}

		@Test
		void shouldCallBuildAttachedItemAsMap() {
			var itemMap = Map.of("key", (Object) "value");
			doReturn(itemMap).when(service).buildItemMapWithExistingBescheidFields(any());

			service.updateBescheidDraft(command);

			verify(service).buildAttachedItemAsMap(command, itemMap);
		}

		@Test
		void shouldCallBuildUpdateAttachedItemCommand() {
			var expectedBodyObject = Map.of("key", (Object) "value");
			doReturn(expectedBodyObject).when(service).buildAttachedItemAsMap(any(), any());

			service.updateBescheidDraft(command);

			verify(service).buildUpdateAttachedItemCommand(bescheidItem, expectedBodyObject);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(updateAttachedItemCommand).when(service).buildUpdateAttachedItemCommand(any(), any());

			service.updateBescheidDraft(command);

			verify(commandService).createAndWaitUntilDone(updateAttachedItemCommand);
		}

		@Test
		void shouldThrowExceptionWhenNotFound() {
			when(remoteService.getBescheid(any())).thenThrow(StatusRuntimeException.class);

			assertThrows(StatusRuntimeException.class, () -> service.updateBescheidDraft(command));
		}

	}

	@Nested
	class TestBuildAttachedItemAsMap {

		private Command command = CommandTestFactory.createBuilder().bodyObject(BescheidItemTestFactory.createBescheidBody()).build();

		@Test
		void shouldSetVorgangId() {
			var map = service.buildAttachedItemAsMap(command, Map.of());

			assertThat(map).containsEntry(BescheidItem.PROPERTY_VORGANG_ID, command.getVorgangId());
		}

		@Test
		void shouldSetClient() {
			var map = service.buildAttachedItemAsMap(command, Map.of());

			assertThat(map).containsEntry(BescheidItem.PROPERTY_CLIENT, BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		}

		@Test
		void shouldSetItemName() {
			var map = service.buildAttachedItemAsMap(command, Map.of());

			assertThat(map).containsEntry(BescheidItem.PROPERTY_ITEM_NAME, AttachedItemService.BESCHEID_ITEM_NAME);
		}

		@Test
		void shouldSetItem() {
			var expectedMap = Map.of("key", (Object) "value");

			var map = service.buildAttachedItemAsMap(command, expectedMap);

			assertThat(map).containsEntry(BescheidItem.PROPERTY_ITEM, expectedMap);
		}
	}

	@Nested
	class TestOverrideAttachedItem {

		private static final String CREATED_RESOURCE = "created-resource";
		@Mock
		private OzgCloudCommand updateItemCommand;

		private Command command = CommandTestFactory.createBuilder().bodyObject(BescheidItemTestFactory.createBescheidBody()).build();

		@BeforeEach
		void init() {
			when(commandService.createAndWaitUntilDone(any())).thenReturn(updateItemCommand);
			when(updateItemCommand.getCreatedResource()).thenReturn(CREATED_RESOURCE);
		}

		@Test
		void shouldCallBuildItemMapWithAllBescheidFields() {
			service.overrideAttachedItem(BescheidItemTestFactory.create(), command);

			verify(service).buildItemMapWithAllBescheidFields(command);
		}

		@Test
		void shouldCallBuildAttachedItemAsMap() {
			var itemMap = Map.of("key", (Object) "value");
			doReturn(itemMap).when(service).buildItemMapWithAllBescheidFields(any());

			service.overrideAttachedItem(BescheidItemTestFactory.create(), command);

			verify(service).buildAttachedItemAsMap(command, itemMap);
		}

		@Test
		void shouldCallBuildUpdateAttachedItemCommand() {
			var bodyObject = Map.of("key", (Object) "value");
			doReturn(bodyObject).when(service).buildAttachedItemAsMap(any(), any());
			var bescheidItem = BescheidItemTestFactory.create();

			service.overrideAttachedItem(bescheidItem, command);

			verify(service).buildUpdateAttachedItemCommand(bescheidItem, bodyObject);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(updateItemCommand).when(service).buildUpdateAttachedItemCommand(any(), any());

			service.overrideAttachedItem(BescheidItemTestFactory.create(), command);

			verify(commandService).createAndWaitUntilDone(updateItemCommand);
		}

		@Test
		void shouldReturnCreatedResource() {
			var result = service.overrideAttachedItem(BescheidItemTestFactory.create(), command);

			assertThat(result).isEqualTo(CREATED_RESOURCE);
		}
	}

	@Nested
	class TestAddValueFromMapIfExists {

		private static final String KEY = "key";

		@Test
		void shouldAddValue() {
			var targetMap = new HashMap<String, Object>();
			var expectedValue = "value";
			Map<String, Object> sourceMap = Map.of(KEY, expectedValue);

			service.addValueFromMapIfExists(sourceMap, KEY, targetMap);

			assertThat(targetMap).containsEntry(KEY, expectedValue);
		}

		@Test
		void shouldNotAddValueIfKeyNotExists() {
			var targetMap = new HashMap<String, Object>();
			var expectedValue = "value";
			Map<String, Object> sourceMap = Map.of(KEY, expectedValue);

			service.addValueFromMapIfExists(sourceMap, "notExists", targetMap);

			assertThat(targetMap).isEmpty();
		}

		@Test
		void shouldAddNullValues() {
			var targetMap = new HashMap<String, Object>();
			Map<String, Object> sourceMap = Collections.singletonMap(KEY, null);

			service.addValueFromMapIfExists(sourceMap, KEY, targetMap);

			assertThat(targetMap).containsEntry(KEY, null);
		}
	}

	@Nested
	class TestBuildBescheidMap {

		private Command command = CommandTestFactory.createBuilder().bodyObject(BescheidItemTestFactory.createBescheidBody()).build();

		@Test
		void shouldSetStatus() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_STATUS, BescheidItem.Status.DRAFT.name());
		}

		@Test
		void shouldSetBeschiedenAm() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_BESCHIEDEN_AM, BescheidItemTestFactory.BESCHEIDEN_AM);
		}

		@Test
		void shouldSetBewilligt() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_BEWILLIGT, true);
		}

		@Test
		void shouldSetBescheidDocument() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_BESCHEID_DOCUMENT, BescheidItemTestFactory.BESCHEID_DOCUMENT);
		}

		@Test
		void shouldSetSendBy() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_SEND_BY, BescheidItemTestFactory.SEND_BY);
		}

		@Test
		void shouldSetNachrichtText() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_NACHRICHT_TEXT, BescheidItemTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldSetNachrichtSubject() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_NACHRICHT_SUBJECT, BescheidItemTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldSetAttachments() {
			var bescheidMap = service.buildItemMapWithExistingBescheidFields(command);

			assertThat(bescheidMap).containsEntry(BescheidItem.FIELD_ATTACHMENTS, List.of(BescheidItemTestFactory.ATTACHMENT));
		}
	}

	@Nested
	class TestBuildUpdateAttachedItemCommand {

		@Test
		void shouldCallVorgangIdMapper() {
			service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), Map.of());

			verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetVorgangId() {
			var expectedVorgangId = OzgCloudVorgangId.from(CommandTestFactory.VORGANG_ID);
			when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

			var result = service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), Map.of());

			assertThat(result.getVorgangId()).isEqualTo(expectedVorgangId);
		}

		@Test
		void shouldCallRelationIdMapper() {
			service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), Map.of());

			verify(commandMapper).mapRelationId(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldSetRelationId() {
			var expectedId = GenericId.from("relationId");
			when(commandMapper.mapRelationId(any())).thenReturn(expectedId);

			var result = service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), Map.of());

			assertThat(result.getRelationId()).isEqualTo(expectedId);
		}

		@Test
		void shouldSetRelationVersion() {
			var result = service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), Map.of());

			assertThat(result.getRelationVersion()).isEqualTo(BescheidItemTestFactory.VERSION);
		}

		@Test
		void shouldSetOrder() {
			var result = service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), Map.of());

			assertThat(result.getOrder()).isEqualTo(AttachedItemService.UPDATE_ATTACHED_ITEM_ORDER);
		}

		@Test
		void shouldSetBodyObject() {
			var bodyObject = Map.of("key", (Object) "value");

			var result = service.buildUpdateAttachedItemCommand(BescheidItemTestFactory.create(), bodyObject);

			assertThat(result.getBodyObject()).containsAllEntriesOf(bodyObject);
		}

	}

	@Nested
	class TestDeleteBescheidDraft {

		@Mock
		private OzgCloudCommand deleteItemOzgCloudCommand;

		private Command command = CommandTestFactory.createBuilder().relationId(BescheidItemTestFactory.ID).build();

		@Test
		void shouldCallGetBescheid() {
			when(remoteService.getBescheid(any())).thenReturn(BescheidItemTestFactory.create());

			deleteBescheidDraft();

			verify(remoteService).getBescheid(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldCallCommandService() {
			doReturn(deleteItemOzgCloudCommand).when(service).buildDeleteItemCommand(any(), any());
			when(remoteService.getBescheid(any())).thenReturn(BescheidItemTestFactory.create());

			deleteBescheidDraft();

			verify(commandService).createAndWaitUntilDone(deleteItemOzgCloudCommand);
		}

		@Test
		void shouldThrowExceptionIfBescheidNotFound() {
			when(remoteService.getBescheid(any())).thenThrow(StatusRuntimeException.class);

			assertThrows(StatusRuntimeException.class, this::deleteBescheidDraft);
		}

		private void deleteBescheidDraft() {
			service.deleteBescheidDraft(command);
		}

		@Nested
		class TestBuildDeleteItemCommand {

			@Test
			void shouldCallCommandMapperWithVorgangId() {
				service.buildDeleteItemCommand(command, BescheidItemTestFactory.create());

				verify(commandMapper).toOzgCloudVorgangId(CommandTestFactory.VORGANG_ID);
			}

			@Test
			void shouldSetVorgangId() {
				var expectedVorgangId = OzgCloudVorgangId.from(CommandTestFactory.VORGANG_ID);
				when(commandMapper.toOzgCloudVorgangId(any())).thenReturn(expectedVorgangId);

				var ozgCloudCommand = service.buildDeleteItemCommand(command, BescheidItemTestFactory.create());

				assertThat(ozgCloudCommand.getVorgangId()).isEqualTo(expectedVorgangId);
			}

			@Test
			void shouldCallCommandMapperWithRelationId() {
				service.buildDeleteItemCommand(command, BescheidItemTestFactory.create());

				verify(commandMapper).mapRelationId(BescheidItemTestFactory.ID);
			}

			@Test
			void shouldSetRelationId() {
				var expectedId = GenericId.from(BescheidItemTestFactory.ID);
				when(commandMapper.mapRelationId(any())).thenReturn(expectedId);

				var ozgCloudCommand = service.buildDeleteItemCommand(command, BescheidItemTestFactory.create());

				assertThat(ozgCloudCommand.getRelationId()).isEqualTo(expectedId);
			}

			@Test
			void shouldSetRelationVersion() {
				var ozgCloudCommand = service.buildDeleteItemCommand(command, BescheidItemTestFactory.create());

				assertThat(ozgCloudCommand.getRelationVersion()).isEqualTo(BescheidItemTestFactory.VERSION);
			}

			@Test
			void shouldSetOrder() {
				var ozgCloudCommand = service.buildDeleteItemCommand(command, BescheidItemTestFactory.create());

				assertThat(ozgCloudCommand.getOrder()).isEqualTo(AttachedItemService.DELETE_ATTACHED_ITEM);
			}
		}
	}

	@Nested
	class TestValidateBescheid {

		@Test
		void shouldFilterDraftStatus() {
			assertDoesNotThrow(() -> service.validateBescheidStatus(BescheidItemTestFactory.create()));
		}

		@Test
		void shouldThrowExceptionIfBescheidIsNotDraft() {
			var bescheidItem = BescheidItemTestFactory.createBuilder().bescheidData(Map.of(BescheidItem.FIELD_STATUS, "otherStatus")).build();

			assertThrows(TechnicalException.class, () -> service.validateBescheidStatus(bescheidItem));
		}

	}

	@Nested
	class TestGetItem {

		@Test
		void shouldCallRemoteService() {
			service.getItem(BescheidItemTestFactory.ID);

			verify(service).getItem(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldReturnValue() {
			var expectedItem = AttachedItemTestFactory.createDocument();
			doReturn(expectedItem).when(remoteService).getItem(any());

			var result = service.getItem(BescheidItemTestFactory.ID);

			assertThat(result).isEqualTo(expectedItem);
		}
	}
}