/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid.attacheditem;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.NullAndEmptySource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.bescheid.GrpcBescheid;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;

class BescheidItemMapperTest {

	@Spy
	@InjectMocks
	private BescheidItemMapper mapper;

	@Mock
	private GrpcObjectMapper grpcObjectMapper;

	@Nested
	class TestFromFindVorgangAttachedItemResponse {

		GrpcFindVorgangAttachedItemResponse grpcResponse = GrpcFindVorgangAttachedItemResponseTestFactory.create();

		@Test
		void shouldCallMapFromVorgangAttachedItem() {
			mapper.fromFindVorgangAttachedItemResponse(grpcResponse).toList();

			verify(mapper).mapFromVorgangAttachedItem(grpcResponse.getVorgangAttachedItems(0));
		}

		@Test
		void shouldReturnBescheidITem() {
			var bescheid = BescheidItemTestFactory.create();
			doReturn(bescheid).when(mapper).mapFromVorgangAttachedItem(any());

			var result = mapper.fromFindVorgangAttachedItemResponse(grpcResponse).toList();

			assertThat(result).first().isEqualTo(bescheid);
		}
	}

	@Nested
	class TestMapFromVorgangAttachedItem {

		@Test
		void shouldSetId() {
			var result = mapper.mapFromVorgangAttachedItem(GrpcVorgangAttachedItemTestFactory.create());

			assertThat(result.getId()).isEqualTo(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldSetVersion() {
			var grpcAttachedItem = GrpcVorgangAttachedItemTestFactory.createBuilder().setVersion(BescheidItemTestFactory.VERSION).build();

			var result = mapper.mapFromVorgangAttachedItem(grpcAttachedItem);

			assertThat(result.getVersion()).isEqualTo(BescheidItemTestFactory.VERSION);
		}

		@Test
		void shouldSetClient() {
			var result = mapper.mapFromVorgangAttachedItem(GrpcVorgangAttachedItemTestFactory.create());

			assertThat(result.getClient()).isEqualTo(BescheidItemTestFactory.CLIENT);
		}

		@Test
		void shouldSetVorgangId() {
			var result = mapper.mapFromVorgangAttachedItem(GrpcVorgangAttachedItemTestFactory.create());

			assertThat(result.getVorgangId()).isEqualTo(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetItemName() {
			var result = mapper.mapFromVorgangAttachedItem(GrpcVorgangAttachedItemTestFactory.create());

			assertThat(result.getItemName()).isEqualTo(AttachedItemService.BESCHEID_ITEM_NAME);
		}

		@Test
		void shouldCallGrpcObjectMapper() {
			var grpcAttachedItem = GrpcVorgangAttachedItemTestFactory.create();

			mapper.mapFromVorgangAttachedItem(grpcAttachedItem);

			verify(grpcObjectMapper).mapFromGrpc(grpcAttachedItem.getItem());
		}

		@Test
		void shouldSetBescheidData() {
			var bescheidData = BescheidItemTestFactory.createBescheidBody();
			when(grpcObjectMapper.mapFromGrpc(any())).thenReturn(bescheidData);

			var result = mapper.mapFromVorgangAttachedItem(GrpcVorgangAttachedItemTestFactory.create());

			assertThat(result.getBescheidData()).isEqualTo(bescheidData);
		}
	}

	@Nested
	class TestToBescheid {

		@Test
		void shouldSetBeschiedenAm() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getBeschiedenAm()).isEqualTo(BescheidItemTestFactory.BESCHEIDEN_AM);
		}

		@Test
		void shouldSetBewilligt() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getBewilligt()).isTrue();
		}

		@Test
		void shouldSetBescheidDocument() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getBescheidDocument()).isEqualTo(BescheidItemTestFactory.BESCHEID_DOCUMENT);
		}

		@Test
		void shouldSetSendBy() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getSendBy()).isEqualTo(BescheidItemTestFactory.SEND_BY);
		}

		@Test
		void shouldSetNachrichtText() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getNachrichtText()).isEqualTo(BescheidItemTestFactory.NACHRICHT_TEXT);
		}

		@Test
		void shouldSetNachrichtSubject() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getNachrichtSubject()).isEqualTo(BescheidItemTestFactory.NACHRICHT_SUBJECT);
		}

		@Test
		void shouldCallToAttachments() {
			var expectedAttachments = List.of(BescheidItemTestFactory.ATTACHMENT);
			var bescheidItem = BescheidItemTestFactory.createBuilder()
					.bescheidData(Map.of(BescheidItem.FIELD_ATTACHMENTS, expectedAttachments)).build();

			mapper.toBescheid(bescheidItem);

			verify(mapper).toAttachments(expectedAttachments);
		}

		@Test
		void shouldNotCallToAttachments() {
			var bescheidItem = BescheidItemTestFactory.createBuilder().bescheidData(Map.of()).build();

			mapper.toBescheid(bescheidItem);

			verify(mapper, never()).toAttachments(any());
		}

		@Test
		void shouldSetAttachments() {
			doReturn(List.of(BescheidItemTestFactory.ATTACHMENT)).when(mapper).toAttachments(any());

			var bescheid = mapBescheid();

			assertThat(bescheid.getAttachmentsList()).containsExactly(BescheidItemTestFactory.ATTACHMENT);
		}

		@Test
		void shouldSetId() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getId()).isEqualTo(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldSetVersion() {
			var bescheid = mapBescheid();

			assertThat(bescheid.getVersion()).isEqualTo(BescheidItemTestFactory.VERSION);
		}

		private GrpcBescheid mapBescheid() {
			return mapper.toBescheid(BescheidItemTestFactory.create());
		}
	}

	@Nested
	class TestToAttachments {

		@Test
		void shouldMapCollection() {
			var result = mapper.toAttachments(List.of(BescheidItemTestFactory.ATTACHMENT));

			assertThat(result).containsExactly(BescheidItemTestFactory.ATTACHMENT);
		}

		@Test
		void shouldMapSingleValue() {
			var result = mapper.toAttachments(BescheidItemTestFactory.ATTACHMENT);

			assertThat(result).containsExactly(BescheidItemTestFactory.ATTACHMENT);
		}

		@DisplayName("should return null")
		@ParameterizedTest(name = "when attachmentsObject is \"{0}\"")
		@NullAndEmptySource
		void shouldReturnNull(String attachmentsObject) {
			var result = mapper.toAttachments(attachmentsObject);

			assertThat(result).isNull();
		}
	}
}