/*
 *
 *  Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 *  Ministerpräsidenten des Landes Schleswig-Holstein
 *  Staatskanzlei
 *  Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 *  Lizenziert unter der EUPL, Version 1.2 oder - sobald
 *  diese von der Europäischen Kommission genehmigt wurden -
 *  Folgeversionen der EUPL ("Lizenz");
 *  Sie dürfen dieses Werk ausschließlich gemäß
 *  dieser Lizenz nutzen.
 *  Eine Kopie der Lizenz finden Sie hier:
 *
 *  https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 *  Sofern nicht durch anwendbare Rechtsvorschriften
 *  gefordert oder in schriftlicher Form vereinbart, wird
 *  die unter der Lizenz verbreitete Software "so wie sie
 *  ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 *  ausdrücklich oder stillschweigend - verbreitet.
 *  Die sprachspezifischen Genehmigungen und Beschränkungen
 *  unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.bescheid.attacheditem;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import org.apache.commons.collections.MapUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.bescheid.BescheidCallContextAttachingInterceptor;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcFindVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItem;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemRequest;
import de.ozgcloud.vorgang.vorgangAttachedItem.GrpcVorgangAttachedItemResponse;
import de.ozgcloud.vorgang.vorgangAttachedItem.VorgangAttachedItemServiceGrpc.VorgangAttachedItemServiceBlockingStub;
import io.grpc.ClientInterceptor;

class VorgangAttachedItemRemoteServiceTest {

	private static final String VORGANG_ID = "vorgang-id";

	@Spy
	@InjectMocks
	private VorgangAttachedItemRemoteService service;

	@Mock
	private VorgangAttachedItemServiceBlockingStub serviceStub;
	@Mock
	private ClientInterceptor bescheidCallContextInterceptor;
	@Mock
	private BescheidItemMapper bescheidItemMapper;
	@Mock
	private AttachedItemMapper attachedItemMapper;

	@Nested
	class TestFindBescheidDraft {

		@Mock
		private GrpcFindVorgangAttachedItemRequest request;

		@BeforeEach
		void init() {
			doReturn(serviceStub).when(service).getServiceStub();
		}

		@Nested
		@DisplayName("by vorgangId")
		class TestGetByVorgangId {

			@Test
			void shouldCallCreateFindRequestBuilder() {
				service.findBescheidDraft(VORGANG_ID);

				verify(service).buildFindRequest(VORGANG_ID);
			}

			@Test
			void shouldCallGetBescheidDraft() {
				doReturn(request).when(service).buildFindRequest(any());

				service.findBescheidDraft(VORGANG_ID);

				verify(service).findBescheidDraft(request);
			}
		}

		@Nested
		@DisplayName("by grpc request")
		class TestByGrpcRequest {

			@Mock
			private GrpcFindVorgangAttachedItemResponse response;

			@Test
			void shouldCallGetServiceStub() {
				service.findBescheidDraft(request);

				verify(service).getServiceStub();
			}

			@Test
			void shouldCallFind() {
				service.findBescheidDraft(request);

				verify(serviceStub).find(request);
			}

			@Test
			void shouldCallMapper() {
				when(serviceStub.find(any())).thenReturn(response);

				service.findBescheidDraft(request);

				verify(bescheidItemMapper).fromFindVorgangAttachedItemResponse(response);
			}

			@Test
			void shouldCallHasStatusDraft() {
				var item = BescheidItemTestFactory.create();
				when(bescheidItemMapper.fromFindVorgangAttachedItemResponse(any())).thenReturn(Stream.of(item));

				service.findBescheidDraft(request);

				verify(service).hasStatusDraft(item);
			}

			@Test
			void shouldFilterDraftItems() {
				var item = BescheidItemTestFactory.create();
				when(bescheidItemMapper.fromFindVorgangAttachedItemResponse(any())).thenReturn(Stream.of(item));
				doReturn(true).when(service).hasStatusDraft(any());

				var result = service.findBescheidDraft(request);

				assertThat(result).contains(item);
			}

			@Test
			void shouldReturnEmptyWhenNotFound() {
				when(bescheidItemMapper.fromFindVorgangAttachedItemResponse(any())).thenReturn(Stream.empty());

				var result = service.findBescheidDraft(request);

				assertThat(result).isEmpty();
			}

			@Test
			void shouldReturnEmptyWhenNotDraft() {
				when(bescheidItemMapper.fromFindVorgangAttachedItemResponse(any())).thenReturn(Stream.of(BescheidItemTestFactory.create()));
				doReturn(false).when(service).hasStatusDraft(any());

				var result = service.findBescheidDraft(request);

				assertThat(result).isEmpty();
			}

			@Test
			void shouldReturnLastBescheid() {
				var item1 = BescheidItemTestFactory.createBuilder().bescheidData(Map.of(
						BescheidItem.FIELD_BESCHIEDEN_AM, "2021-01-01")).build();
				var item2 = BescheidItemTestFactory.create();
				when(bescheidItemMapper.fromFindVorgangAttachedItemResponse(any())).thenReturn(Stream.of(item1, item2));
				doReturn(true).when(service).hasStatusDraft(any());

				var result = service.findBescheidDraft(request);

				assertThat(getBeschiedenAm(result)).isEqualTo(BescheidItemTestFactory.BESCHEIDEN_AM);
			}

			private String getBeschiedenAm(Optional<BescheidItem> item) {
				assertThat(item).isPresent();
				return MapUtils.getString(item.get().getBescheidData(), BescheidItem.FIELD_BESCHIEDEN_AM);
			}
		}

	}

	@Nested
	class TestBuildFindRequest {

		@Test
		void shouldBuildRequest() {
			var request = buildFindRequest();

			assertThat(request.getVorgangId()).isEqualTo(CommandTestFactory.VORGANG_ID);
		}

		@Test
		void shouldSetItemName() {
			var request = buildFindRequest();

			assertThat(request.getItemName()).isEqualTo(AttachedItemService.BESCHEID_ITEM_NAME);
		}

		@Test
		void shouldSetClient() {
			var request = buildFindRequest();

			assertThat(request.getClient()).isEqualTo(BescheidCallContextAttachingInterceptor.BESCHEID_MANAGER_CLIENT_NAME);
		}

		private GrpcFindVorgangAttachedItemRequest buildFindRequest() {
			return service.buildFindRequest(CommandTestFactory.VORGANG_ID);
		}
	}

	@Nested
	class TestHasStatusDraft {

		@Test
		void shouldReturnTrueIfDraft() {
			var item = BescheidItemTestFactory.createBuilder().bescheidData(Map.of(BescheidItem.FIELD_STATUS, BescheidItem.Status.DRAFT.name()))
					.build();

			var result = service.hasStatusDraft(item);

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseIfNotDraft() {
			var item = BescheidItemTestFactory.createBuilder().bescheidData(Map.of(BescheidItem.FIELD_STATUS, BescheidItem.Status.BESCHEID.name()))
					.build();

			var result = service.hasStatusDraft(item);

			assertThat(result).isFalse();
		}
	}

	@Nested
	class TestGetBescheid {

		@Mock
		private GrpcVorgangAttachedItemRequest grpcVorgangAttachedItemRequest;
		@Mock
		private GrpcVorgangAttachedItemResponse grpcVorgangAttachedItemResponse;
		@Mock
		private GrpcVorgangAttachedItem grpcVorgangAttachedItem;

		@BeforeEach
		void init() {
			when(serviceStub.getById(any())).thenReturn(grpcVorgangAttachedItemResponse);
			doReturn(serviceStub).when(service).getServiceStub();
		}

		@Test
		void shouldCallGetServiceStab() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getBescheid();

			verify(service).getServiceStub();
		}

		@Test
		void shouldCallBuildRequest() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getBescheid();

			verify(service).buildGetByIdRequest(BescheidItemTestFactory.ID);
		}

		@Test
		void shouldCallGetById() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);
			doReturn(grpcVorgangAttachedItemRequest).when(service).buildGetByIdRequest(any());

			getBescheid();

			verify(serviceStub).getById(grpcVorgangAttachedItemRequest);
		}

		@Test
		void shouldCallMapper() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getBescheid();

			verify(bescheidItemMapper).mapFromVorgangAttachedItem(grpcVorgangAttachedItem);
		}

		@Test
		void shouldReturnFoundBescheid() {
			var expectedBescheid = BescheidItemTestFactory.create();
			when(bescheidItemMapper.mapFromVorgangAttachedItem(any())).thenReturn(expectedBescheid);
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			var result = getBescheid();

			assertThat(result).isEqualTo(expectedBescheid);
		}

		private BescheidItem getBescheid() {
			return service.getBescheid(BescheidItemTestFactory.ID);
		}
	}

	@Nested
	class TestGetItem {

		@Mock
		private GrpcVorgangAttachedItemRequest grpcVorgangAttachedItemRequest;
		@Mock
		private GrpcVorgangAttachedItemResponse grpcVorgangAttachedItemResponse;
		@Mock
		private GrpcVorgangAttachedItem grpcVorgangAttachedItem;

		@BeforeEach
		void init() {
			when(serviceStub.getById(any())).thenReturn(grpcVorgangAttachedItemResponse);
			doReturn(serviceStub).when(service).getServiceStub();
		}

		@Test
		void shouldCallGetServiceStab() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(service).getServiceStub();
		}

		@Test
		void shouldCallBuildRequest() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(service).buildGetByIdRequest(AttachedItemTestFactory.ID);
		}

		@Test
		void shouldCallGetById() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);
			doReturn(grpcVorgangAttachedItemRequest).when(service).buildGetByIdRequest(any());

			getItem();

			verify(serviceStub).getById(grpcVorgangAttachedItemRequest);
		}

		@Test
		void shouldCallMapper() {
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			getItem();

			verify(attachedItemMapper).mapFromVorgangAttachedItem(grpcVorgangAttachedItem);
		}

		@Test
		void shouldReturnFoundBescheid() {
			var expectedItem = AttachedItemTestFactory.createDocument();
			when(attachedItemMapper.mapFromVorgangAttachedItem(any())).thenReturn(expectedItem);
			when(grpcVorgangAttachedItemResponse.getVorgangAttachedItem()).thenReturn(grpcVorgangAttachedItem);

			var result = getItem();

			assertThat(result).isEqualTo(expectedItem);
		}

		private AttachedItem getItem() {
			return service.getItem(BescheidItemTestFactory.ID);
		}
	}

	@Nested
	class TestBuildGetByIdRequest {

		@Test
		void shouldSetBescheidId() {
			var request = service.buildGetByIdRequest(BescheidItemTestFactory.ID);

			assertThat(request.getId()).isEqualTo(BescheidItemTestFactory.ID);
		}
	}

	@Nested
	class TestGetServiceStub {

		@Test
		void shouldCallWithInterceptors() {
			service.getServiceStub();

			verify(serviceStub).withInterceptors(bescheidCallContextInterceptor);
		}
	}
}