/*
 * Copyright (c) 2024. Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.nachrichten.antragraum;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import jakarta.annotation.PostConstruct;

import org.opensaml.core.xml.schema.XSString;
import org.opensaml.saml.saml2.core.AttributeStatement;
import org.opensaml.saml.saml2.core.EncryptedAssertion;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.saml.saml2.encryption.EncryptedElementTypeEncryptedKeyResolver;
import org.opensaml.security.credential.Credential;
import org.opensaml.security.credential.CredentialSupport;
import org.opensaml.xmlsec.encryption.support.ChainingEncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.EncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.InlineEncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.SimpleRetrievalMethodEncryptedKeyResolver;
import org.opensaml.xmlsec.keyinfo.KeyInfoCredentialResolver;
import org.opensaml.xmlsec.keyinfo.impl.CollectionKeyInfoCredentialResolver;
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty;
import org.springframework.security.saml2.Saml2Exception;
import org.springframework.stereotype.Service;

import lombok.RequiredArgsConstructor;

@Service
@RequiredArgsConstructor
@ConditionalOnProperty(AntragraumProperties.PROPERTY_ANTRAGSRAUM_URL)
class Saml2Decrypter {
	private final BayernIdSamlConfiguration configuration;

	public static final String LEGACY_POSTKORB_HANDLE_KEY = "legacyPostkorbHandle";
	private Decrypter decrypter;
	private static final EncryptedKeyResolver encryptedKeyResolver = new ChainingEncryptedKeyResolver(
			Arrays.asList(new InlineEncryptedKeyResolver(), new EncryptedElementTypeEncryptedKeyResolver(),
					new SimpleRetrievalMethodEncryptedKeyResolver()));

	@PostConstruct
	void init() {
		Collection<Credential> credentials = new ArrayList<>();
		var decryptionX509Credential = configuration.getDecryptionCredential();

		Credential cred = CredentialSupport.getSimpleCredential(decryptionX509Credential.getCertificate(), decryptionX509Credential.getPrivateKey());
		credentials.add(cred);

		KeyInfoCredentialResolver resolver = new CollectionKeyInfoCredentialResolver(credentials);
		var setupDecrypter = new Decrypter(null, resolver, encryptedKeyResolver);
		setupDecrypter.setRootInNewDocument(true);

		decrypter = setupDecrypter;
	}

	void decryptResponseElements(Response response) {
		for (EncryptedAssertion encryptedAssertion : response.getEncryptedAssertions()) {
			try {
				var assertion = decrypter.decrypt(encryptedAssertion);
				response.getAssertions().add(assertion);
			} catch (Exception ex) {
				throw new Saml2Exception(ex);
			}
		}
	}

	String decryptPostfachId(Response response) {
		decryptResponseElements(response);

		var samlAssertion = response.getAssertions().get(0);
		var statements = (AttributeStatement) samlAssertion.getStatements().get(1);
		var attributes = statements.getAttributes();
		var postfachIdOptional = attributes.stream().filter(attribute -> LEGACY_POSTKORB_HANDLE_KEY.equals(attribute.getFriendlyName())).findFirst();

		return postfachIdOptional.map(postfachIdAttribute -> {
			var values = postfachIdAttribute.getAttributeValues();
			return ((XSString) values.get(0)).getValue();
		}).orElseThrow();

	}
}
