/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import java.time.ZonedDateTime;
import java.util.EnumSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Stream;

import jakarta.validation.Valid;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import de.ozgcloud.nachrichten.antragraum.AntragraumService;
import de.ozgcloud.nachrichten.attributes.ClientAttributeService;
import de.ozgcloud.nachrichten.info.InfoManagerService;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.ReplyOption;
import de.ozgcloud.nachrichten.postfach.osi.OsiPostfachServerProcessException;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
@Validated
class PostfachService {

	private static final Predicate<PostfachNachricht> IS_FROM_HUMAN_USER = nachricht -> !StringUtils.startsWith(nachricht.getCreatedBy(), "system");
	private static final Set<String> POSTFACH_TYPES_WITH_ANTRAGSRAUM = Set.of("BayernId");

	@Autowired(required = false)
	private PostfachRemoteService postfachRemoteService;

	private static final Set<ReplyOption> REPLY_POSSIBLE_OPTION = EnumSet.of(ReplyOption.POSSIBLE, ReplyOption.MANDATORY);

	@Autowired
	private PostfachNachrichtMapper mapper;

	@Autowired
	private PersistPostfachNachrichtService persistingService;
	@Autowired
	private ClientAttributeService clientAttributeService;

	@Autowired
	private ApplicationEventPublisher publisher;

	@Autowired
	private CurrentUserService userService;

	@Autowired
	private Optional<InfoManagerService> infomanagerService;
	@Autowired
	private Optional<AntragraumService> antragraumService;

	public void saveDraft(String vorgangId, PostfachNachricht nachricht) {
		persistMail(userService.getUser().getUserId(),
				nachricht.toBuilder().vorgangId(vorgangId)
						.createdBy(userService.getUser().getUserId().orElse(null))
						.build());
	}

	public void sendMail(String commandId, String userId, @Valid PostfachNachricht mail) {
		var sendResponse = handleSendMail(commandId, mail);
		persistSentMail(userId, addMailSentInformation(mail, sendResponse));
	}

	PostfachNachricht addMailSentInformation(PostfachNachricht mail, SendPostfachNachrichtResponse sendResponse) {
		return mail.toBuilder().sentAt(ZonedDateTime.now()).sentSuccessful(sendResponse.isSentSuccessful())
				.messageCode(sendResponse.getMessageCode().getMessageCode())
				.build();
	}

	void persistSentMail(@NonNull String userId, PostfachNachricht mail) {
		persistMail(Optional.of(userId), mail.toBuilder()
				.direction(Direction.OUT)
				.createdAt(ZonedDateTime.now().withNano(0)).createdBy(userId)
				.build());

	}

	public Optional<Map<String, Object>> findById(String nachrichtId) {
		return persistingService.findById(nachrichtId);
	}

	public Stream<Map<String, Object>> findByVorgang(String vorgangId) {
		return persistingService.findByVorgangAsMap(vorgangId);
	}

	public void fetchAndPersistReplies() {
		ifPostfachConfigured(() -> postfachRemoteService.getAllMessages().forEach(this::persistReceivedMail));
	}

	private void persistReceivedMail(PostfachNachricht nachricht) {
		persistMail(Optional.empty(), nachricht);
		clientAttributeService.setHasNewPostfachNachricht(nachricht.getVorgangId());

		postfachRemoteService.deleteMessage(nachricht.getMessageId());
	}

	void persistMail(Optional<String> userId, PostfachNachricht mail) {
		persistingService.persistNachricht(userId, mail);

		if (IS_FROM_HUMAN_USER.test(mail)) {
			clientAttributeService.setHasPostfachNachricht(mail.getVorgangId());
		}
	}

	public void resendMail(String commandId, String postfachMailId) {
		PostfachNachricht nachricht = mapper.fromMapToPostfachMail(persistingService.getById(postfachMailId));

		var sendResponse = handleSendMail(commandId, nachricht);
		patchMail(nachricht.getId(), createResendPatchMap(sendResponse));
	}

	SendPostfachNachrichtResponse handleSendMail(String commandId, PostfachNachricht mail) {
		try {
			var processedPostfachNachricht = processForAntragsraum(mail).orElse(mail);
			doSendMail(processedPostfachNachricht);

			publishMailSentEvent(commandId);
			return buildSendNachrichtResponse(true, PostfachMessageCode.SEND_SUCCESSFUL_MESSAGE_CODE);
		} catch (OsiPostfachServerProcessException e) {
			return proceedwithWarnException(commandId, e);

		} catch (PostfachException e) {
			return proceedWithErrorException(commandId, e);
		}
	}

	Optional<PostfachNachricht> processForAntragsraum(PostfachNachricht mail) {
		if (isNotifyAntragsraum(mail.getReplyOption())) {
			infomanagerService.ifPresent(infoManager -> infoManager.notifyInfoManager(mail));
			return adjustMail(mail);
		}
		return Optional.empty();
	}

	boolean isNotifyAntragsraum(ReplyOption replyOption) {
		return REPLY_POSSIBLE_OPTION.contains(replyOption) && isPostfachWithAntragsraum(postfachRemoteService);
	}

	Optional<PostfachNachricht> adjustMail(PostfachNachricht nachricht) {
		return antragraumService.map(antragraum -> nachricht.toBuilder().mailBody(antragraum.getUserNotificationText()).build());
	}

	SendPostfachNachrichtResponse proceedwithWarnException(String commandId, OsiPostfachServerProcessException e) {
		LOG.warn(e.getMessage(), e);
		return proceedWithException(commandId, e);
	}

	SendPostfachNachrichtResponse proceedWithErrorException(String commandId, PostfachException e) {
		LOG.error(e.getMessage(), e);
		return proceedWithException(commandId, e);
	}

	private SendPostfachNachrichtResponse proceedWithException(String commandId, PostfachException e) {
		publishMailSentFailedEvent(commandId, e.getMessage());
		return buildSendNachrichtResponse(false, e.getMessageCode());

	}

	void doSendMail(PostfachNachricht nachricht) {
		ifPostfachConfigured(() -> postfachRemoteService.sendMessage(nachricht));
	}

	private void publishMailSentEvent(String commandId) {
		publisher.publishEvent(new PostfachMailSentEvent(commandId));
	}

	public void publishMailSentFailedEvent(String commandId, String message) {
		publisher.publishEvent(new PostfachMailSentFailedEvent(commandId, message));
	}

	private SendPostfachNachrichtResponse buildSendNachrichtResponse(boolean sentSuccesful, PostfachMessageCode messageCode) {
		return SendPostfachNachrichtResponse.builder().sentSuccessful(sentSuccesful).messageCode(messageCode).build();
	}

	Map<String, Object> createResendPatchMap(SendPostfachNachrichtResponse sendResponse) {
		return Map.of(PostfachNachricht.FIELD_SENT_AT, ZonedDateTime.now().withNano(0).toString(),
				PostfachNachricht.FIELD_SENT_SUCCESSFUL, sendResponse.isSentSuccessful(),
				PostfachNachricht.FIELD_MESSAGE_CODE, sendResponse.getMessageCode().getMessageCode());
	}

	private void patchMail(String postfachMailId, Map<String, Object> propertyMap) {
		persistingService.patch(postfachMailId, propertyMap);
	}

	public boolean isPostfachConfigured() {
		return postfachRemoteService != null;
	}

	public Stream<Postfach> getPostfachs() {
		return isPostfachConfigured() ? Stream.of(buildPostfach(postfachRemoteService)) : Stream.empty();
	}

	Postfach buildPostfach(PostfachRemoteService postfachRemoteService) {
		return Postfach.builder().type(postfachRemoteService.getPostfachType()).isReplyAllowed(isReplyAllowed(postfachRemoteService)).build();
	}

	boolean isReplyAllowed(PostfachRemoteService postfachRemoteService) {
		return isPostfachWithAntragsraum(postfachRemoteService) || postfachRemoteService.isReplyAllowed();
	}

	boolean isPostfachWithAntragsraum(PostfachRemoteService postfachRemoteService) {
		return POSTFACH_TYPES_WITH_ANTRAGSRAUM.contains(postfachRemoteService.getPostfachType()) && infomanagerService.isPresent();
	}

	private void ifPostfachConfigured(Runnable block) {
		if (!isPostfachConfigured()) {
			throw new NotConfiguredException();
		}
		block.run();
	}

}