/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.nachrichten.postfach;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;

import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import lombok.SneakyThrows;

class PostfachNachrichtMapperTest {

	@InjectMocks
	private PostfachNachrichtMapper mapper = Mappers.getMapper(PostfachNachrichtMapper.class);
	@Mock
	private GrpcObjectMapper grpcObjectMapper;

	@DisplayName("From map")
	@Nested
	class TestFromMap {

		@DisplayName("with existing postfachAddress")
		@Nested
		class TestWithExistingPostfachAddress {

			@BeforeEach
			void mockGrpcObjectMapper() {
				when(grpcObjectMapper.fromMap(any())).thenReturn(GrpcPostfachAddressTestFactory.createIdentifier());
			}

			@Test
			void shouldFillGrpcMail() {
				var grpcMail = fromMap(PostfachNachrichtTestFactory.asMap());

				assertThat(grpcMail).usingRecursiveComparison().isEqualTo(GrpcPostfachMailTestFactory.create());
			}

			@Test
			void shouldCallGrpcObjectMapper() {
				fromMap(getMapWithoutProperty(PostfachNachricht.FIELD_SENT_SUCCESSFUL));

				verify(grpcObjectMapper).fromMap(any());
			}

			@Test
			void shouldIgnoreSentSuccessful() {
				var grpcMail = fromMap(getMapWithoutProperty(PostfachNachricht.FIELD_SENT_SUCCESSFUL));

				assertThat(grpcMail.getSentSuccessful()).isFalse();
			}

			@Test
			void shouldIgnoreSentAt() {
				var grpcMail = fromMap(getMapWithoutProperty(PostfachNachricht.FIELD_SENT_AT));

				assertThat(grpcMail.getSentAt()).isEmpty();
			}

			@ParameterizedTest
			@ValueSource(strings = PostfachNachricht.FIELD_POSTFACH_ID)
			void shouldMapDefaultOnNonExistingValues(String fieldName) {
				var mapped = fromMap(getMapWithoutProperty(fieldName));

				var fieldValue = getAttributeFromGrpcObject(mapped, fieldName);
				assertThat((String) fieldValue).isEmpty();
			}
		}

		@DisplayName("with missing postfachAddress")
		@Nested
		class TestWithMissinPostfachAddress {

			@Test
			void shouldNotThrowException() {
				assertDoesNotThrow(() -> fromMap(getMapWithoutProperty(PostfachNachricht.POSTFACH_ADDRESS_FIELD)));
			}
		}

		private GrpcPostfachMail fromMap(Map<String, Object> map) {
			return mapper.fromMap(map);
		}
	}

	@DisplayName("From map to postfachMail")
	@Nested
	class TestFromMapToPostfachMail {

		@Test
		void shouldMap() {
			var mail = fromMapToPostfachMail(PostfachNachrichtTestFactory.asMap());

			assertThat(mail).usingRecursiveComparison().isEqualTo(PostfachNachrichtTestFactory.create());
		}

		@DisplayName("with missing postfachAddress")
		@Nested
		class TestWithMissinPostfachAddress {

			@Test
			void shouldMapAsNull() {
				var mail = fromMapToPostfachMail(getMapWithoutProperty(PostfachNachricht.POSTFACH_ADDRESS_FIELD));

				assertThat(mail.getPostfachAddress()).isNull();
			}

			@Test
			void shouldNotThrowExceptionOnMissingPostfachAddress() {
				assertDoesNotThrow(() -> fromMapToPostfachMail(getMapWithoutProperty(PostfachNachricht.POSTFACH_ADDRESS_FIELD)));
			}
		}

		@ParameterizedTest
		@ValueSource(strings = PostfachNachricht.FIELD_POSTFACH_ID)
		void shouldMapDefaultOnNonExistingValues(String fieldName) {
			var mapped = fromMapToPostfachMail(getMapWithoutProperty(fieldName));

			var fieldValue = getAttributeFromObject(mapped, fieldName);
			assertThat((String) fieldValue).isEmpty();
		}

		private PostfachNachricht fromMapToPostfachMail(Map<String, Object> map) {
			return mapper.fromMapToPostfachMail(map);
		}
	}

	@DisplayName("From grpc mail")
	@Nested
	class TestFromGrpcMail {

		@Test
		void shouldMapFields() {
			var mapped = fromGrpcMail(GrpcPostfachMailTestFactory.create());

			assertThat(mapped).usingRecursiveComparison()
					.ignoringFields("createdAt", "createdBy", "direction", "messageId", "sentAt", "sentSuccessful", "memoizedHashCode")
					.isEqualTo(PostfachNachrichtTestFactory.create());
		}

		@Test
		void shouldHandleEmptyPostfachId() {
			var postfachMailWithoutPostfachId = GrpcPostfachMailTestFactory.createBuilder().clearPostfachId().build();

			assertDoesNotThrow(() -> fromGrpcMail(postfachMailWithoutPostfachId));
		}

		private PostfachNachricht fromGrpcMail(GrpcPostfachMail postfachMail) {
			return mapper.fromGrpcMail(postfachMail);
		}
	}

	private Map<String, Object> getMapWithoutProperty(String entryKey) {
		var map = PostfachNachrichtTestFactory.asMap();
		map.remove(entryKey);
		return map;
	}

	private Object getAttributeFromGrpcObject(Object obj, String name) {
		return getAttributeFrom(obj, name + "_");
	}

	@SneakyThrows
	private Object getAttributeFromObject(Object obj, String name) {
		return getAttributeFrom(obj, name);
	}

	@SneakyThrows
	private Object getAttributeFrom(Object obj, String name) {
		var clazz = obj.getClass();
		var field = clazz.getDeclaredField(name);
		field.setAccessible(true);

		return field.get(obj);
	}
}
