/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.notification.antragsteller;

import java.util.Objects;
import java.util.Optional;
import java.util.function.Predicate;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.event.EventListener;
import org.springframework.stereotype.Component;

import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.notification.postfach.PostfachService;
import de.ozgcloud.notification.vorgang.Vorgang;
import de.ozgcloud.notification.vorgang.VorgangId;
import de.ozgcloud.notification.vorgang.VorgangService;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
public class AntragstellerNotificationEventListener {

	static final String FS_FORMENGINE_NAME = "FormSolutions";
	static final String POSTFACH_NOT_CONFIGURED_MESSAGE = "Postfach is not configured. Antragsteller notifications will not be sent.";
	private static final Predicate<Vorgang> EXISTS_POSTFACH_ID = vorgang -> StringUtils.isNoneBlank(vorgang.getPostfachId());
	private static final Predicate<Vorgang> EXISTS_POSTFACH_ADDRESS = vorgang -> Objects.nonNull(vorgang.getPostfachAddress());
	private static final Predicate<Vorgang> IS_FORMSOLUTIONS_FORM_ENGINE = vorgang -> StringUtils.equals(vorgang.getFormEngineName(),
			FS_FORMENGINE_NAME);
	private static final Predicate<Vorgang> SHOULD_SEND_NOTIFICATION = vorgang -> //
			(EXISTS_POSTFACH_ID.test(vorgang) || EXISTS_POSTFACH_ADDRESS.test(vorgang)) && IS_FORMSOLUTIONS_FORM_ENGINE.test(vorgang);

	@Autowired
	private AntragstellerNotificationService service;
	@Autowired
	private VorgangService vorgangService;
	@Autowired
	private PostfachService postfachService;

	private Optional<Boolean> isPostfachConfigured = Optional.empty();

	@EventListener
	public void onVorgangCreated(VorgangCreatedEvent event) {
		if (isPostfachConfigured()) {
			sendIfRequired(vorgangService.getVorgang(VorgangId.from(event.getSource())));
		} else {
			LOG.debug(POSTFACH_NOT_CONFIGURED_MESSAGE);
		}
	}

	boolean isPostfachConfigured() {
		return isPostfachConfigured.orElseGet(this::fetchIsPostfachConfigured);
	}

	synchronized boolean fetchIsPostfachConfigured() {
		return isPostfachConfigured.orElseGet(() -> {
			var isConfigured = postfachService.isPostfachConfigured();
			if (!isConfigured) {
				LOG.warn(POSTFACH_NOT_CONFIGURED_MESSAGE);
			}
			isPostfachConfigured = Optional.of(isConfigured);
			return isConfigured;
		});
	}

	private void sendIfRequired(Vorgang vorgang) {
		if (SHOULD_SEND_NOTIFICATION.test(vorgang)) {
			service.sendNotification(vorgang);
		}
	}
}
