/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.notification.antragsteller;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.notification.command.CommandService;
import de.ozgcloud.notification.postfach.PostfachAddressTestFactory;
import de.ozgcloud.notification.postfach.PostfachNachricht;
import de.ozgcloud.notification.postfach.PostfachNachrichtTestFactory;
import de.ozgcloud.notification.vorgang.VorgangTestFactory;

class AntragstellerNotificationServiceTest {

	@Spy
	@InjectMocks
	private AntragstellerNotificationService service;
	@Mock
	private CommandService commandService;

	@Captor
	private ArgumentCaptor<PostfachNachricht> nachrichtCaptor;

	@DisplayName("Send notification")
	@Nested
	class TestSendNotification {

		@Test
		void shouldCallCommandService() {
			service.sendNotification(VorgangTestFactory.create());

			verify(commandService).createSendPostfachNachrichtCommand(nachrichtCaptor.capture());

			assertThat(nachrichtCaptor.getValue())
					.usingRecursiveComparison().ignoringFields("mailBody", "subject")
					.isEqualTo(PostfachNachrichtTestFactory.create());
		}

		@Nested
		class BuildPostfachNachricht {
			@Test
			void shouldFormatMessage() {
				service.buildPostfachNachricht(VorgangTestFactory.create());

				verify(service).formatMessage(any());
			}

			@Test
			void shouldFormSubject() {
				service.buildPostfachNachricht(VorgangTestFactory.create());

				verify(service).formatSubject(any());
			}

			@Test
			void shouldHaveVorgangId() {
				var nachricht = service.buildPostfachNachricht(VorgangTestFactory.create());

				assertThat(nachricht.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			@Test
			void shouldHavePostfachId() {
				var nachricht = service.buildPostfachNachricht(VorgangTestFactory.create());

				assertThat(nachricht.getPostfachId()).isEqualTo(VorgangTestFactory.POSTFACH_ID);
			}
			
			@Test
			void shouldHavePostfachAddress() {
				var nachricht = service.buildPostfachNachricht(VorgangTestFactory.create());

				assertThat(nachricht.getPostfachAddress()).usingRecursiveComparison().isEqualTo(PostfachAddressTestFactory.create());
			}
		}

		@Nested
		class FormatMessage {

			@Test
			void shouldContaineDateTimeInMEZ() {
				var vorgang = VorgangTestFactory.createBuilder().createdAt(ZonedDateTime.parse("2001-01-01T05:00:00Z")).build();

				var message = service.formatMessage(vorgang);

				assertThat(message).contains("01.01.2001 06:00:00");
			}

			@Test
			void shouldContaineDateTimeInMESZ() {
				var vorgang = VorgangTestFactory.createBuilder().createdAt(ZonedDateTime.parse("2001-08-01T05:00:00Z")).build();

				var message = service.formatMessage(vorgang);

				assertThat(message).contains("01.08.2001 07:00:00");
			}

			@Test
			void shouldContainVorgangNummer() {
				var message = service.formatMessage(VorgangTestFactory.create());

				assertThat(message).contains(VorgangTestFactory.VORGANG_NUMMER);
			}
		}

	}

}
