/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.notification.user;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.notification.email.EmailRemoteService;
import de.ozgcloud.notification.email.UserEmail;
import de.ozgcloud.notification.email.UserEmailTestFactory;
import de.ozgcloud.notification.vorgang.Vorgang;
import de.ozgcloud.notification.vorgang.VorgangTestFactory;

class UserNotificationServiceTest {

	@Spy
	@InjectMocks
	private UserNotificationService service;

	@Mock
	private EmailRemoteService emailRemoteService;
	@Mock
	private UserRemoteService userRemoteService;

	@DisplayName("Send notification")
	@Nested
	class TestSendNotification {

		@DisplayName("with valid orgaId")
		@Nested
		class TestWithValidOrgaId {

			@BeforeEach
			void init() {
				when(userRemoteService.getRecipients(VorgangTestFactory.ORGANISATIONS_EINHEIT_ID)).thenReturn(List.of(RecipientTestFactory.create()));
				doNothing().when(emailRemoteService).sendEmail(any(UserEmail.class));
			}

			@Test
			void shouldGetRecipients() {
				service.sendNotification(VorgangTestFactory.create());

				verify(userRemoteService).getRecipients(VorgangTestFactory.ORGANISATIONS_EINHEIT_ID);
			}

			@Test
			void shouldSendEmail() {
				var email = UserEmailTestFactory.create();
				doReturn(email).when(service).buildUserEmail(anyList(), anyString());

				service.sendNotification(VorgangTestFactory.create());

				verify(emailRemoteService).sendEmail(email);
			}

			@Test
			void shouldSetSubject() {
				service.sendNotification(VorgangTestFactory.create());

				verify(service).buildUserEmail(anyList(), eq("Neuer Vorgang: " + VorgangTestFactory.VORGANG_NAME));

			}
		}

		@DisplayName("with null as orgaId")
		@Nested
		class TestWithInvalidOrgaId {

			private Vorgang vorgang = VorgangTestFactory.createBuilder().organisationseinheitenId(null).build();

			@Test
			void shouldNotCallUserRemoteService() {
				service.sendNotification(vorgang);

				verify(userRemoteService, never()).getRecipients(anyString());
			}

			@Test
			void shouldNotCallEmailRemoteServie() {
				service.sendNotification(vorgang);

				verify(emailRemoteService, never()).sendEmail(any(UserEmail.class));
			}
		}

		@Test
		void shouldThrowExceptionWhenVorgangNull() {
			assertThatThrownBy(() -> service.sendNotification(null))
					.withFailMessage("organisationsEinheitId cannot be null.")
					.isInstanceOf(IllegalArgumentException.class);
		}
	}
}