/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.attached_item;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Stream;

import jakarta.validation.Valid;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import de.ozgcloud.command.Command;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import lombok.NonNull;

@Service
@Validated
public class VorgangAttachedItemService {

	static final boolean NOT_DELETED = false;

	@Autowired
	private VorgangAttachedItemRepository repository;
	@Autowired
	private ApplicationEventPublisher publisher;

	public VorgangAttachedItem create(@NonNull String commandId, @Valid VorgangAttachedItem item) {
		var saved = repository.save(prepareForCreation(item));

		publisher.publishEvent(new VorgangAttachedItemCreatedEvent(commandId, saved.getId()));

		return saved;
	}

	VorgangAttachedItem prepareForCreation(VorgangAttachedItem item) {
		return item.toBuilder()
				.id(null)
				.version(0)
				.item(removeId(item.getItem()))
				.build();
	}

	private Map<String, Object> removeId(@NonNull Map<String, Object> item) {
		var map = new HashMap<String, Object>(item);
		map.remove(VorgangAttachedItem.FIELDNAME_ID);
		return Collections.unmodifiableMap(map);
	}

	public void update(String commandId, VorgangAttachedItem item) {
		repository.patch(item.getId(), item.getVersion(), item.getItem());

		publisher.publishEvent(new VorgangAttachedItemUpdatedEvent(commandId));
	}

	public void forceUpdate(String commandId, VorgangAttachedItem item) {
		repository.forcePatch(item.getId(), item.getItem());

		publisher.publishEvent(new VorgangAttachedItemUpdatedEvent(commandId));
	}

	public Optional<VorgangAttachedItem> findById(String itemId) {
		return repository.findByIdAndDeleted(itemId, NOT_DELETED).map(this::addIdToItem);
	}

	public Stream<VorgangAttachedItem> find(String vorgangId, Optional<String> client, Optional<String> itemName) {
		if (hasVorgangIdOnly(client, itemName)) {
			return repository.findByVorgangIdAndDeleted(vorgangId, NOT_DELETED).map(this::addIdToItem);
		} else if (isEmpty(itemName) && client.isPresent()) {
			return repository.findByVorgangIdAndClientAndDeleted(vorgangId, client.get(), NOT_DELETED).map(this::addIdToItem);
		} else if (isEmpty(client)) {
			return repository.findByVorgangIdAndItemNameAndDeleted(vorgangId, itemName.get(), NOT_DELETED).map(this::addIdToItem);
		} else {
			return repository.findByVorgangIdAndClientAndItemNameAndDeleted(vorgangId, client.get(), itemName.get(), NOT_DELETED)
					.map(this::addIdToItem);
		}
	}

	private boolean isEmpty(Optional<String> client) {
		return client.isEmpty() || client.get().isEmpty();
	}

	private boolean hasVorgangIdOnly(Optional<String> client, Optional<String> itemName) {
		return isEmpty(itemName) && isEmpty(client);
	}

	public void forcePatch(String itemId, Map<String, Object> propertyMap) {
		repository.forcePatch(itemId, propertyMap);
	}

	public void patch(Command command, Map<String, Object> body) {
		repository.patch(command.getRelationId(), command.getRelationVersion(), body);

		publisher.publishEvent(new VorgangAttachedItemUpdatedEvent(command.getId()));
	}

	public VorgangAttachedItem getById(String id) {
		var attachedItem = repository.findByIdAndDeleted(id, NOT_DELETED).orElseThrow(() -> new NotFoundException(VorgangAttachedItem.class, id));

		return addIdToItem(attachedItem);
	}

	VorgangAttachedItem addIdToItem(VorgangAttachedItem attachedItem) {
		var itemMap = new HashMap<>(attachedItem.getItem());
		itemMap.put(VorgangAttachedItem.FIELDNAME_ID, attachedItem.getId());
		return attachedItem.toBuilder().item(Collections.unmodifiableMap(itemMap)).build();
	}

	public void markAsDeleteByIdAndVersion(String itemId, long version) {
		repository.updateDeleted(true, itemId, version);
	}

	public void unmarkAsDeleteByIdAndVersion(String itemId, long version) {
		repository.updateDeleted(false, itemId, version);
	}

	public void deleteByVorgangId(String vorgangId) {
		repository.deleteByVorgangId(vorgangId);
	}

}