/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.callcontext.CallContext;
import de.ozgcloud.vorgang.callcontext.CallContextUser;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.callcontext.User;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Service
@Validated
@Log4j2
public class CommandService {

	@Autowired
	private CommandRepository repository;
	@Autowired
	private ApplicationEventPublisher publisher;

	@Autowired
	private CurrentUserService userService;

	public Command createCommand(CreateCommandRequest request) {
		var persisted = saveCommand(buildCommand(request));
		publisher.publishEvent(new CommandCreatedEvent(persisted));
		return persisted;
	}

	private Command buildCommand(CreateCommandRequest request) {
		var builder = PersistedCommand.builder()
				.id(UUID.randomUUID().toString())
				.vorgangId(request.getVorgangId())
				.order(request.getOrder())
				.status(CommandStatus.PENDING)
				.createdAt(ZonedDateTime.now())
				.relationId(request.getRelationId())
				.relationVersion(request.getRelationVersion())
				.redirectRequest(request.getRedirectRequest())
				.body(request.getBody())
				.bodyObject(request.getBodyObject());

		builder = addUserIfExists(builder, Optional.ofNullable(request.getCallContext()).map(CallContext::getUser), userService.findUser());
		return builder.build();
	}

	private PersistedCommand.PersistedCommandBuilder addUserIfExists(PersistedCommand.PersistedCommandBuilder builder, Optional<User> user,
			Optional<CallContextUser> contextUser) {
		if (contextUser.isPresent()) {
			return builder.createdBy(contextUser.get().getUserId().orElse(null)).createdByName(contextUser.get().getUserName().orElse(null));
		}

		return user.map(u -> builder.createdBy(u.getId()).createdByName(u.getName())).orElse(builder);
	}

	Command saveCommand(Command command) {
		repository.save(command);

		return command;
	}

	public void setCommandFinished(String commandId, long relationVersion) {
		repository.updateCommandStatusAndVersion(commandId, CommandStatus.FINISHED, relationVersion);
	}

	public Optional<Command> findCommand(String commandId) {
		return repository.getById(commandId);
	}

	public Stream<Command> findFinishedVorgangLoeschenCommandsOlderThen(@NonNull ZonedDateTime createdAfter) {
		return repository.findCommands(Order.VORGANG_LOESCHEN, CommandStatus.FINISHED, createdAfter);
	}

	public void setCommandFinished(String commandId) {
		repository.finishCommand(commandId);
	}

	public void setCommandFinished(String commandId, String createdResource) {
		repository.finishCommand(commandId, createdResource);
	}

	public void setCommandError(String id, String errorMessage) {
		repository.setErrorMessage(id, errorMessage);
	}

	public void setCommandRevoked(String commandId) {
		updateCommandStatus(commandId, CommandStatus.REVOKED);
	}

	public void setCommandRevokePending(final String commandId) {
		updateCommandStatus(commandId, CommandStatus.REVOKE_PENDING);

		publishRevokeCommandEvent(commandId);
	}

	void publishRevokeCommandEvent(final String commandId) {
		var commandOptional = repository.getById(commandId);
		commandOptional.ifPresentOrElse(command -> {
			if (MapUtils.isEmpty(((PersistedCommand) command).getPreviousState())) {
				handleError(commandId, "Command %s can not be revoked because it has no previous state.");
			}
			publisher.publishEvent(new RevokeCommandEvent(command));
		}, () -> handleError(commandId, "No Command with id %s found."));
	}

	private void handleError(String commandId, String messageTemplate) {
		var message = messageTemplate.formatted(commandId);
		LOG.error(message);
		publisher.publishEvent(new CommandFailedEvent(commandId, message));
		throw new TechnicalException(message);
	}

	void updateCommandStatus(String commandId, CommandStatus status) {
		repository.updateCommandStatus(commandId, status);
	}

	public boolean existsPendingCommands(String vorgangId) {
		return repository.existsPendingCommands(vorgangId);
	}

	public List<Command> getPendingCommands(String relationId) {
		return repository.getPendingCommands(relationId);
	}

	public Stream<Command> findCommands(String vorgangId, Collection<CommandStatus> status, Optional<Order> order) {
		return repository.findCommands(vorgangId, status, order).stream();
	}

	public void setPreviousState(String commandId, Map<String, Object> previousState) {
		repository.patch(commandId, previousState);
	}

	public void deleteAllByVorgang(String vorgangId) {
		repository.deleteAllByVorgang(vorgangId);
	}

	public boolean existsCommandWithUserId(String userId) {
		return repository.existsCommandWithUserId(userId);
	}
}