/*
 * Copyright (c) 2024.
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URLConnection;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Base64;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.stream.Stream;

import org.apache.commons.lang3.StringUtils;
import org.apache.http.entity.ContentType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.nachrichten.postfach.AttachmentFile;
import de.ozgcloud.nachrichten.postfach.PersistPostfachNachrichtService;
import de.ozgcloud.nachrichten.postfach.PostfachAddress;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.Direction;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht.ReplyOption;
import de.ozgcloud.nachrichten.postfach.StringBasedIdentifier;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemMapper;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemService;
import de.ozgcloud.vorgang.callcontext.CallContext;
import de.ozgcloud.vorgang.callcontext.User;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.files.OzgFile;
import de.ozgcloud.vorgang.files.UploadedFilesReference;
import jakarta.activation.MimetypesFileTypeMap;
import lombok.extern.log4j.Log4j2;

/**
 * @deprecated Temporally replacement for using GRPC Api by MailService
 */
@Deprecated
@Service
@Log4j2
class PersistPostfachNachrichtByCommandService implements PersistPostfachNachrichtService {

	static final String CLIENT = "OzgCloud_NachrichtenManager";
	static final String ITEM_NAME = "PostfachMail";
	static final String ATTACHMENT_NAME = "PostfachAttachment";

	@Autowired
	private CommandService service;
	@Autowired
	private VorgangAttachedItemService attachedItemService;
	@Autowired
	private FileService fileService;

	// Dummy implementation
	// TODO OZG-4768 replace it
	private final Map<String, Collection<PostfachNachricht>> answers = new HashMap<>();

	@Override
	public void persistNachricht(Optional<String> userId, PostfachNachricht nachricht) {
		var builder = CreateCommandRequest.builder()
				.callContext(buildCallContext(userId))
				.vorgangId(nachricht.getVorgangId());

		var request = StringUtils.isBlank(nachricht.getId()) ? buildPersistCommand(builder, nachricht) : buildUpdateCommand(builder, nachricht);

		service.createCommand(request);
	}

	CreateCommandRequest buildPersistCommand(CreateCommandRequest.CreateCommandRequestBuilder builder, PostfachNachricht nachricht) {
		return builder.order(Order.CREATE_ATTACHED_ITEM.name())
				.relationId(nachricht.getVorgangId())
				.bodyObject(buildItem(nachricht)).build();
	}

	CreateCommandRequest buildUpdateCommand(CreateCommandRequest.CreateCommandRequestBuilder builder, PostfachNachricht nachricht) {
		return builder.order(Order.UPDATE_ATTACHED_ITEM.name())
				.relationId(nachricht.getId())
				.bodyObject(buildUpdateItem(nachricht)).build();
	}

	private CallContext buildCallContext(Optional<String> userId) {
		var builder = CallContext.builder();
		userId.map(id -> User.builder().id(id).build()).ifPresent(builder::user);

		return builder.build();
	}

	private Map<String, Object> buildItem(PostfachNachricht nachricht) {
		return Map.of(
				VorgangAttachedItemMapper.PROPERTY_CLIENT, CLIENT,
				VorgangAttachedItemMapper.PROPERTY_VORGANG_ID, nachricht.getVorgangId(),
				VorgangAttachedItemMapper.PROPERTY_ITEM_NAME, ITEM_NAME,
				VorgangAttachedItemMapper.PROPERTY_ITEM, buildNachrichtMap(nachricht));

	}

	private Map<String, Object> buildUpdateItem(PostfachNachricht nachricht) {
		return Map.of(
				VorgangAttachedItemMapper.PROPERTY_CLIENT, CLIENT,
				VorgangAttachedItemMapper.PROPERTY_ITEM, buildNachrichtMap(nachricht));

	}

	Map<String, Object> buildNachrichtMap(PostfachNachricht nachricht) {
		var result = new HashMap<>(Map.of(
				PostfachNachricht.FIELD_VORGANG_ID, nachricht.getVorgangId(),
				PostfachNachricht.FIELD_CREATED_AT, nachricht.getCreatedAt().format(DateTimeFormatter.ISO_DATE_TIME),
				PostfachNachricht.FIELD_DIRECTION, nachricht.getDirection().name(),
				PostfachNachricht.FIELD_REPLY_OPTION, nachricht.getReplyOption().name(),
				PostfachNachricht.FIELD_SUBJECT, nachricht.getSubject(),
				PostfachNachricht.FIELD_MAIL_BODY, nachricht.getMailBody(),
				PostfachNachricht.FIELD_ATTACHMENTS, nachricht.getAttachments()));

		putIfNonNull(result, PostfachNachricht.FIELD_POSTFACH_ID, nachricht.getPostfachId());
		putIfNonNull(result, PostfachNachricht.FIELD_MESSAGE_ID, nachricht.getMessageId());
		putIfNonNull(result, PostfachNachricht.FIELD_CREATED_BY, nachricht.getCreatedBy());
		putIfNonNull(result, PostfachNachricht.FIELD_SENT_SUCCESSFUL, nachricht.getSentSuccessful());
		putIfNonNull(result, PostfachNachricht.FIELD_MESSAGE_CODE, nachricht.getMessageCode());
		putDateIfNonNull(result, PostfachNachricht.FIELD_SENT_AT, nachricht.getSentAt());

		Optional.ofNullable(nachricht.getPostfachAddress()).map(this::buildPostfachAddressMap)
				.ifPresent(postfachMaiMap -> result.put(PostfachNachricht.POSTFACH_ADDRESS_FIELD, postfachMaiMap));

		return Collections.unmodifiableMap(result);
	}

	private Map<String, Object> putIfNonNull(Map<String, Object> mapIn, String key, Object value) {
		if (Objects.nonNull(value)) {
			mapIn.put(key, value);
		}
		return mapIn;
	}

	private Map<String, Object> putDateIfNonNull(Map<String, Object> mapIn, String key, Object value) {
		if (Objects.nonNull(value)) {
			mapIn.put(key, ((ZonedDateTime) value).format(DateTimeFormatter.ISO_DATE_TIME));
		}
		return mapIn;
	}

	private Map<String, Object> buildPostfachAddressMap(PostfachAddress postfachAddress) {
		var resultMap = new HashMap<String, Object>();
		resultMap.put(PostfachAddress.IDENTIFIER_FIELD, buildPostfachAddressIdentifierMap(postfachAddress));
		resultMap.put(PostfachAddress.VERSION_FIELD, postfachAddress.getVersion());
		resultMap.put(PostfachAddress.TYPE_FIELD, postfachAddress.getType());
		Optional.ofNullable(postfachAddress.getServiceKontoType()).ifPresentOrElse(
				serviceKontoType -> resultMap.put(PostfachAddress.SERVICEKONTO_TYPE_FIELD, serviceKontoType),
				() -> LOG.warn("ServiceKontoType is null"));
		return Collections.unmodifiableMap(resultMap);
	}

	private Map<String, Object> buildPostfachAddressIdentifierMap(PostfachAddress postfachAddress) {
		var identifier = (StringBasedIdentifier) postfachAddress.getIdentifier();
		return Map.of(PostfachNachricht.FIELD_POSTFACH_ID, identifier.getPostfachId());
	}

	@Override
	public Optional<Map<String, Object>> findById(String postfachNachrichtId) {
		return attachedItemService.findById(postfachNachrichtId).map(VorgangAttachedItem::getItem);
	}

	@Override
	public Stream<Map<String, Object>> findByVorgangAsMap(String vorgangId) {
		return attachedItemService.find(vorgangId, Optional.of(CLIENT), Optional.of(ITEM_NAME)).map(VorgangAttachedItem::getItem);
	}

	@Override
	public void patch(String postfachNachrichtId, Map<String, Object> propertyMap) {
		attachedItemService.forcePatch(postfachNachrichtId, propertyMap);
	}

	@Override
	public Map<String, Object> getById(String postfachNachrichtId) {
		return attachedItemService.getById(postfachNachrichtId).getItem();
	}

	@Override
	public String persistAttachment(String vorgangId, AttachmentFile attachment) {
		var contentType = getTypeByFile(attachment);
		try (var content = attachment.getContent()) {
			var decContent = Base64.getDecoder().decode(content.readAllBytes());

			return fileService.uploadFileStream(
					createUploadedFilesReference(vorgangId),
					createOzgFile(attachment.getName(), contentType, decContent.length),
					Optional.empty(),
					new ByteArrayInputStream(decContent)).get(10, TimeUnit.MINUTES).toString();
		} catch (ExecutionException | TimeoutException e) {
			throw new TechnicalException(e.getMessage(), e);
		} catch (InterruptedException e) {
			Thread.currentThread().interrupt();
			throw new TechnicalException(e.getMessage(), e);
		} catch (IOException e) {
			throw new TechnicalException("Can not read attached file", e);
		}
	}

	UploadedFilesReference createUploadedFilesReference(String vorgangId) {
		return UploadedFilesReference.builder().vorgangId(vorgangId).client(CLIENT).name(ATTACHMENT_NAME).build();
	}

	OzgFile createOzgFile(String fileName, String contentType, long size) {
		return OzgFile.builder().name(fileName).contentType(contentType).size(size).build();
	}

	String getTypeByFile(AttachmentFile attachmentFile) {
		var fileNameMap = URLConnection.getFileNameMap();

		return Optional.ofNullable(fileNameMap.getContentTypeFor(attachmentFile.getName())).orElseGet(() -> getTypeByContent(attachmentFile));
	}

	private String getTypeByContent(AttachmentFile attachmentFile) {
		try (var contentStream = attachmentFile.getContent()) {
			return Optional.ofNullable(URLConnection.guessContentTypeFromStream(contentStream))
					.orElseGet(() -> getByMimeTypes(attachmentFile.getName()));
		} catch (IOException e) {
			LOG.warn("IO-Exception while guessing content type", e);
		}
		return ContentType.APPLICATION_OCTET_STREAM.toString();
	}

	// uses map file: src/main/resources/mime.types
	private String getByMimeTypes(String fileName) {
		var fileTypeMap = new MimetypesFileTypeMap();

		return fileTypeMap.getContentType(fileName);
	}

	@Override
	public Stream<PostfachNachricht> findRueckfragen(String postfachId) {
		// Dummy Implementierung
		// TODO OZG-5132 do real implementation
		return Stream.of(PostfachNachricht.builder()
				.attachments(Collections.emptyList())
				.createdAt(ZonedDateTime.now())
				.sentAt(ZonedDateTime.now())
				.vorgangId("VORGANG_DUMMY_URI")
				.createdBy("Klaus")
				.direction(Direction.OUT)
				.replyOption(ReplyOption.POSSIBLE)
				.mailBody("Lorem ipsum dolres est")
				.id(UUID.randomUUID().toString())
				.subject("Test")
				.build());
	}

	@Override
	public String persistAnswer(String rueckfrageId, PostfachNachricht answer) {
		// Dummy implementation
		// TODO OZG-4768 replace it by throwing Exception here (use gRPC for persisting)
		var id = UUID.randomUUID().toString();
		answer = answer.toBuilder().messageId(id).build();

		var collection = Optional.ofNullable(answers.get(rueckfrageId)).orElseGet(() -> {
			var newSet = new HashSet<PostfachNachricht>();
			answers.put(rueckfrageId, newSet);
			return newSet;
		});
		collection.add(answer);
		return id;
	}

	@Override
	public Stream<PostfachNachricht> findAnswers(String rueckfrageId) {
		// Dummy implementation
		// TODO OZG-4768 replace it by throwing Exception here (use gRPC for loading)
		return Optional.ofNullable(answers.get(rueckfrageId))
				.orElse(Collections.emptySet()).stream();
	}

}