/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.security;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Optional;

import org.bson.Document;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.aggregation.Aggregation;
import org.springframework.data.mongodb.core.aggregation.AggregationOperation;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.stereotype.Repository;

import de.ozgcloud.command.Command;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.common.db.CriteriaUtil;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.vorgang.Vorgang;

@Repository
class PolicyRepository {

	private static final String VORANG_ID_AS_OBJECT_ID_FIELD = """
			{"$addFields" : {"vorgangObjecId" : {"$toObjectId": "$vorgangId"}}}
			""";
	private static final String VORANG_ID_FROM_FILE_AS_OBJECT_ID_FIELD = """
			{"$addFields" : {"vorgangObjecId" : {"$toObjectId": "$metadata.vorgangId"}}}
			""";
	private static final AggregationOperation ADD_VORGANG_OBJECT_ID_FIELD = context -> org.bson.Document.parse(VORANG_ID_AS_OBJECT_ID_FIELD);

	private static final AggregationOperation ADD_VORGANG_FROM_FILE = context -> org.bson.Document.parse(VORANG_ID_FROM_FILE_AS_OBJECT_ID_FIELD);

	public static final String FIELD_ORGANISATIONSEINHEIT_ID = "organisationseinheitenId";
	public static final String FIELD_VORGANG = "vorgang";
	public static final String FIELD_VORGANG_ORGANISATIONSEINHEIT = FIELD_VORGANG + "." + Vorgang.FIELD_ORGANISATIONSEINHEIT;

	@Autowired
	private MongoTemplate template;

	public boolean existsByCommandId(String commandId, Collection<String> organisationEinheitenIds) {
		var document = executeAggregation(buildAggregation(matchId(commandId), ADD_VORGANG_OBJECT_ID_FIELD), Command.class).orElseThrow(
				() -> new NotFoundException(Command.class, commandId));
		return evaluateOrganisationseinheitId(document, organisationEinheitenIds);
	}

	public boolean existsByFileId(String fileId, Collection<String> organisationEinheitenIds) {
		var resultDocument = executeAggregation(buildAggregation(matchId(fileId), ADD_VORGANG_FROM_FILE), "fs.files");
		return resultDocument.map(document -> evaluateOrganisationseinheitId(document, organisationEinheitenIds)).orElse(false);
	}

	private AggregationOperation matchId(String id) {
		return Aggregation.match(new Criteria(Vorgang.MONGODB_FIELDNAME_ID).is(id));
	}

	public boolean existsByVorgangAttachedItem(String vorgangAttachedItemId, Collection<String> organisationEinheitenIds) {
		var result = executeAggregation(buildAggregation(matchVorgangAttachedItemId(vorgangAttachedItemId), ADD_VORGANG_OBJECT_ID_FIELD),
				VorgangAttachedItem.class).orElseThrow(() -> new NotFoundException(VorgangAttachedItem.class, vorgangAttachedItemId));
		return evaluateOrganisationseinheitId(result, organisationEinheitenIds);
	}

	private AggregationOperation matchVorgangAttachedItemId(String id) {
		return Aggregation.match(CriteriaUtil.isId(id).and(VorgangAttachedItem.FIELDNAME_IS_DELETED).is(false));
	}

	private Aggregation buildAggregation(AggregationOperation idAggregation, AggregationOperation operation) {
		return Aggregation.newAggregation(Arrays.asList(
				idAggregation,
				operation,
				lookupVorgang(),
				Aggregation.project().and(FIELD_VORGANG_ORGANISATIONSEINHEIT).as(FIELD_ORGANISATIONSEINHEIT_ID)
		));
	}

	private AggregationOperation lookupVorgang() {
		return Aggregation.lookup(Vorgang.COLLECTION_NAME, "vorgangObjecId", "_id", FIELD_VORGANG);
	}

	private Optional<Document> executeAggregation(Aggregation aggregation, Class<?> inputType) {
		var results = template.aggregate(aggregation, inputType, Document.class).getRawResults();
		return getResult(results);
	}

	private Optional<Document> executeAggregation(Aggregation aggregation, String collectionName) {
		var results = template.aggregate(aggregation, collectionName, Document.class).getRawResults();
		return getResult(results);
	}

	Optional<Document> getResult(Document aggregationResult) {
		return aggregationResult.getList("results", Document.class).stream().findFirst();
	}

	boolean evaluateOrganisationseinheitId(Document result, Collection<String> organisationEinheitenIds) {
		var nestedList = result.getList(FIELD_ORGANISATIONSEINHEIT_ID, Object.class);
		if (nestedList.isEmpty()) {
			return false;
		}
		var organisationseinheitenId = nestedList.get(0);
		if (organisationseinheitenId instanceof List<?> orgaIdList) {
			return orgaIdList.stream().map(String::valueOf).anyMatch(organisationEinheitenIds::contains);
		}
		return organisationEinheitenIds.contains(String.valueOf(organisationseinheitenId));
	}

}