/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Autowired;

import com.google.protobuf.ByteString;

import de.ozgcloud.nachrichten.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceImplBase;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcBinaryFilesRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.extern.log4j.Log4j2;
import net.devh.boot.grpc.server.service.GrpcService;

@Log4j2
@GrpcService
public class GrpcBinaryFileService extends BinaryFileServiceImplBase {

	// FIXME use client from security context
	static final String CLIENT = "Alfa"; // currently single client is hardcoded

	@Autowired
	private FileService service;
	@Autowired
	private GrpcOzgFileMapper fileMapper;
	@Autowired
	private PolicyService policyService;

	@Override
	public void findBinaryFilesMetaData(GrpcBinaryFilesRequest request, StreamObserver<GrpcFindFilesResponse> responseObserver) {
		var found = service.findFilesMetaData(mapFileIds(request.getFileIdList())).toList();

		policyService.checkPermission(found.stream().map(OzgFile::getVorgangId).collect(Collectors.toSet()));

		responseObserver.onNext(buildResponseForOzgFiles(found));
		responseObserver.onCompleted();
	}

	public GrpcFindFilesResponse buildResponseForOzgFiles(List<OzgFile> files) {
		var builder = GrpcFindFilesResponse.newBuilder();

		files.stream().map(file -> fileMapper.map(file)).forEach(builder::addFile);
		return builder.build();
	}

	private List<FileId> mapFileIds(List<String> fileIdList) {
		return fileIdList.stream().map(FileId::from).toList();
	}

	@Override
	public StreamObserver<GrpcUploadBinaryFileRequest> uploadBinaryFileAsStream(StreamObserver<GrpcUploadBinaryFileResponse> responseObserver) {
		return new UploadStreamObserver(responseObserver, service, policyService);
	}

	static GrpcUploadBinaryFileResponse buildResponse(FileId fileId) {
		return GrpcUploadBinaryFileResponse.newBuilder().setFileId(fileId.toString()).build();
	}

	@Override
	public void getBinaryFileContent(GrpcGetBinaryFileDataRequest request, StreamObserver<GrpcGetBinaryFileDataResponse> responseObserver) {
		policyService.checkPermissionByFile(request.getFileId());

		var fileContent = service.getUploadedFileStream(FileId.from(request.getFileId()));

		downloadFile(responseObserver, new BufferedInputStream(fileContent, FileService.CHUNK_SIZE));
	}

	void downloadFile(final StreamObserver<GrpcGetBinaryFileDataResponse> streamObserver, final InputStream fileStream) {
		AtomicBoolean doSendFile = new AtomicBoolean(true);
		final CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver = (CallStreamObserver<GrpcGetBinaryFileDataResponse>) streamObserver;
		callObserver.setOnReadyHandler(() -> sendChunks(callObserver, fileStream, doSendFile));
	}

	void sendChunks(CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver, InputStream fileStream, AtomicBoolean doSendFile) {
		try {
			// TODO remove additional flow control when alfa side has been fixed see
			// OZG-2631
			if (doSendFile.get()) {
				int size = sendNextChunk(callObserver, fileStream);

				if (size < FileService.CHUNK_SIZE) {
					handleFileEndReached(callObserver, fileStream, doSendFile);
				}
			}
		} catch (Exception e) {
			handleException(fileStream, e);
		}
	}

	int sendNextChunk(CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver, InputStream fileStream)
			throws IOException {
		var bytes = new byte[FileService.CHUNK_SIZE];
		int size = fileStream.read(bytes);

		if (size > 0) {
			callObserver.onNext(buildChunkResponse(bytes, size));
		}

		return size;
	}

	private void handleFileEndReached(CallStreamObserver<GrpcGetBinaryFileDataResponse> callObserver, InputStream fileStream,
			AtomicBoolean doSendFile) {
		IOUtils.closeQuietly(fileStream);
		callObserver.onCompleted();
		doSendFile.getAndSet(false);

	}

	private void handleException(InputStream fileStream, Exception e) {
		LOG.error("Fehler beim Senden des Datei-Chunks", e);

		IOUtils.closeQuietly(fileStream);

		throw new TechnicalException("Fehler beim Senden des Datei-Chunks", e);
	}

	GrpcGetBinaryFileDataResponse buildChunkResponse(byte[] bytes, int size) {
		return GrpcGetBinaryFileDataResponse.newBuilder().setFileContent(ByteString.copyFrom(bytes, 0, size)).build();
	}
}