/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.bescheid;

import static de.ozgcloud.bescheid.BescheidEventListener.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.concurrent.TimeUnit;

import org.awaitility.Awaitility;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.test.annotation.DirtiesContext;

import com.mongodb.client.gridfs.model.GridFSFile;

import de.ozgcloud.bescheid.attacheditem.AttachedItemService;
import de.ozgcloud.bescheid.attacheditem.BescheidItem;
import de.ozgcloud.bescheid.common.callcontext.CallContextUser;
import de.ozgcloud.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.bescheid.common.callcontext.UserProfile;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.document.BescheidDocumentCreatedEvent;
import de.ozgcloud.document.Document;
import de.ozgcloud.vorgang.VorgangManagerServerApplication;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@SpringBootTest(classes = VorgangManagerServerApplication.class, properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.ozgcloud-command-manager.address=in-process:test",
		"grpc.client.vorgang-manager.address=in-process:test",
		"ozgcloud.feature.bescheid.enable-dummy-document-processor=true",
})
@DataITCase
@DirtiesContext
class BescheidEventListenerITCase {

	@Autowired
	private ApplicationEventPublisher eventPublisher;
	@SpyBean
	private TestEventListener bescheidDocumentCreatedEventListener;

	@Autowired
	private MongoOperations mongoOperations;
	@Autowired
	private GridFsTemplate gridFsTemplate;

	@MockBean
	private CurrentUserService currentUserService;
	@Mock
	private CallContextUser callContextUser;
	@Mock
	private UserProfile userProfile;

	@Captor
	private ArgumentCaptor<BescheidDocumentCreatedEvent> bescheidDocumentCreatedEventCaptor;

	@BeforeEach
	void init() {
		mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);
		mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
		when(currentUserService.getUser()).thenReturn(callContextUser);
		when(currentUserService.getUserProfile()).thenReturn(userProfile);
	}

	@Test
	void shouldCreateBescheidDocument() {
		var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
		var bescheidItemId = mongoOperations.save(createBescheidAttachedItem(vorgangId)).getId();

		eventPublisher.publishEvent(buildCommandCreatedEvent(vorgangId, bescheidItemId));

		Awaitility.await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
			verify(bescheidDocumentCreatedEventListener).onBescheidDocumentCreated(bescheidDocumentCreatedEventCaptor.capture());
			assertThat(getBescheidDocument()).isNotNull();
		});
	}

	@Test
	void shouldCreateDocumentFile() {
		var vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build()).getId();
		var bescheidItemId = mongoOperations.save(createBescheidAttachedItem(vorgangId)).getId();

		eventPublisher.publishEvent(buildCommandCreatedEvent(vorgangId, bescheidItemId));

		Awaitility.await().atMost(60, TimeUnit.SECONDS).untilAsserted(() -> {
			verify(bescheidDocumentCreatedEventListener).onBescheidDocumentCreated(bescheidDocumentCreatedEventCaptor.capture());
			assertThat(getBescheidDocumentFile()).isNotNull();
		});
	}

	private VorgangAttachedItem createBescheidAttachedItem(String vorgangId) {
		return VorgangAttachedItemTestFactory.createBuilder()
				.id(null)
				.version(0L)
				.vorgangId(vorgangId)
				.itemName(AttachedItemService.BESCHEID_ITEM_NAME)
				.item(Map.of(BescheidItem.FIELD_STATUS, BescheidItem.Status.DRAFT.name()))
				.build();
	}

	private CommandCreatedEvent buildCommandCreatedEvent(String vorgangId, String bescheidItemId) {
		var command = CommandTestFactory.createBuilder()
				.vorgangId(vorgangId)
				.relationId(bescheidItemId)
				.order(BescheidEventListener.CREATE_BESCHEID_DOCUMENT_ORDER)
				.bodyObject(Map.of(BESCHEID_VOM_BODYKEY, "2024-01-01",
						GENEHMIGT_BODYKEY, true))
				.build();
		return new CommandCreatedEvent(command);
	}

	private VorgangAttachedItem getBescheidDocument() {
		var documentId = bescheidDocumentCreatedEventCaptor.getValue().getCreatedResource();
		return mongoOperations.findById(documentId, VorgangAttachedItem.class);
	}

	private GridFSFile getBescheidDocumentFile() {
		var fileId = getBescheidDocument().getItem().get(Document.FIELD_DOCUMENT_FILE).toString();
		return gridFsTemplate.findOne(Query.query(Criteria.where("_id").is(fileId)));
	}
}
