package de.ozgcloud.vorgang.command;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.mockito.Mockito.*;

import java.util.concurrent.TimeUnit;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeReadPermitted;
import de.ozgcloud.vorgang.common.grpc.GrpcObjectMapper;
import de.ozgcloud.vorgang.grpc.command.CommandServiceGrpc.CommandServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.command.GrpcCreateCommandRequest;
import de.ozgcloud.vorgang.grpc.command.GrpcOrder;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=test",
		"grpc.client.inProcess.address=in-process:test"
})
@DirtiesContext
@DataITCase
@WithMockUser
class GrpcCommandServiceITCase {

	@GrpcClient("inProcess")
	private CommandServiceBlockingStub serviceBlockingStub;

	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	private ClientAttributeReadPermitted readIsPermitted;

	private GrpcObjectMapper grpcObjectMapper = Mappers.getMapper(GrpcObjectMapper.class);

	@Nested
	class TestCreateCommand {

		private GrpcCreateCommandRequest request;

		@BeforeEach
		void prepareDatabase() {
			mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
			mongoOperations.dropCollection(Command.COLLECTION_NAME);
			mongoOperations.dropCollection(VorgangAttachedItem.COLLECTION_NAME);

			var vorgang = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).build());

			request = GrpcCreateCommandRequestTestFactory.createBuilder()
					.setVorgangId(vorgang.getId()).setRelationId(vorgang.getId())
					.setOrder(GrpcOrder.CREATE_ATTACHED_ITEM)
					.setBodyObj(grpcObjectMapper.fromMap(VorgangAttachedItemTestFactory.asMap())).build();

			when(readIsPermitted.test(any())).thenReturn(true);
		}

		@Test
		void shouldPersistCommand() {
			serviceBlockingStub.createCommand(request);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
				var commands = mongoOperations.findAll(Command.class);

				assertThat(commands).hasSize(1).first().satisfies(command -> {
					assertThat(command.getBodyObject()).isNotNull();
					assertThat(command.getBodyObject()).containsEntry("itemName", VorgangAttachedItemTestFactory.ITEM_NAME);
				});
			});
		}

		@Test
		void shouldExecuteCommand() {
			serviceBlockingStub.createCommand(request);

			await().atMost(5, TimeUnit.SECONDS).untilAsserted(() -> {
				var commands = mongoOperations.findAll(Command.class);

				assertThat(commands).hasSize(1).first().extracting(Command::getStatus).isEqualTo(CommandStatus.FINISHED);
			});
		}
	}
}
