/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.common.search;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.util.List;
import java.util.Optional;
import java.util.UUID;

import org.apache.commons.lang3.StringUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.elasticsearch.core.ElasticsearchOperations;
import org.springframework.test.context.junit.jupiter.DisabledIf;

import de.ozgcloud.command.Command;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.vorgang.AntragstellerTestFactory;
import de.ozgcloud.vorgang.vorgang.FilterCriteria;
import de.ozgcloud.vorgang.vorgang.FilterCriteriaTestFactory;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequest;
import de.ozgcloud.vorgang.vorgang.FindVorgangRequestTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang.Status;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

@DisabledIf(expression = "${elasticTests.disabled:false}", reason = "Wegen hohem Speicherverbrauch des testcontainers")
@SearchITCase
public class SearchServiceITCase {

	@Autowired
	private SearchService searchService;

	@Autowired
	private SearchVorgangRepository repository;

	@Autowired
	private ElasticsearchOperations elasticsearchOperations;

	@MockBean
	private CommandService commandService;

	@Nested
	class TestAddVorgang {
		@BeforeEach
		void init() {
			elasticsearchOperations.indexOps(IndexedVorgang.class).delete();
		}

		@Test
		void shouldAdd() throws IOException {
			searchService.addVorgang(VorgangTestFactory.create());

			Iterable<IndexedVorgang> vorgangs = repository.findAll();
			assertThat(vorgangs).isNotEmpty();
		}
	}

	@Nested
	class TestUpdateVorgang {
		@BeforeEach
		void init() {
			Command cmd = CommandTestFactory.createBuilder().build();
			when(commandService.findCommand(CommandTestFactory.ID)).thenReturn(Optional.of(cmd));

			elasticsearchOperations.indexOps(IndexedVorgang.class).delete();

			searchService.addVorgang(VorgangTestFactory.create());
		}

		@Test
		void shouldUpdate() throws IOException {
			var vorgang = VorgangTestFactory.createBuilder().status(Status.ABGESCHLOSSEN).build();
			searchService.updateVorgang(vorgang);

			Iterable<IndexedVorgang> vorgangs = repository.findAll();
			assertThat(vorgangs).isNotEmpty().hasSize(1);
			assertThat(vorgangs.iterator().next().getStatus()).isEqualTo(Status.ABGESCHLOSSEN.name());
		}
	}

	@DisplayName("Find")
	@Nested
	class TestFind {

		@BeforeEach
		void init() {
			var indexOperations = elasticsearchOperations.indexOps(IndexedVorgang.class);
			indexOperations.delete();
			indexOperations.create();
			indexOperations.putMapping(IndexedVorgang.class);
		}

		@DisplayName("with hyphen")
		@Nested
		class TestWithHyphen {

			@DisplayName("in vorgangNummer")
			@Test
			void shouldFindInVorgangNummer() {
				var vorgangNummer = "Vorgang-Nummer";
				elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().vorgangNummer(vorgangNummer).build());
				elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

				var vorgangHeaders = searchService.find(
						FindVorgangRequestTestFactory.createBuilder().searchBy(vorgangNummer).build()).getContent();

				assertThat(vorgangHeaders).hasSize(1).first().extracting("nummer").isEqualTo(vorgangNummer);
			}

			@DisplayName("in vorgangName")
			@Test
			void shouldFindInVorgangName() {
				var vorgangName = "VorgangName";
				elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().vorgangName(vorgangName).build());
				elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

				var vorgangHeaders = searchService.find(
						FindVorgangRequestTestFactory.createBuilder().searchBy(vorgangName).build()).getContent();

				assertThat(vorgangHeaders).hasSize(1).first().extracting("name").isEqualTo(vorgangName);

			}

			@DisplayName("in aktenzeichen")
			@Test
			void shouldFindInAktenzeichen() {
				var aktenzeichen = "Akten-zeichen";
				elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().aktenzeichen(aktenzeichen).build());
				elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

				var vorgangHeaders = searchService.find(
						FindVorgangRequestTestFactory.createBuilder().searchBy("kten").build()).getContent();

				assertThat(vorgangHeaders).hasSize(1).first().extracting("aktenzeichen").isEqualTo(aktenzeichen);
			}

			@DisplayName("in antragstellerName")
			@Test
			void shouldFindInAntragstellerName() {
				elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().antragstellerName("Antragsteller-Name").build());
				repository.save(IndexedVorgangTestFactory.createBuilder().vorgangId("123").build());
				elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

				var vorgangHeaders = searchService.find(FindVorgangRequestTestFactory.createBuilder().searchBy("steller-na").build()).getContent();

				assertThat(vorgangHeaders).hasSize(1);
			}

			@DisplayName("in antragstellerVorname")
			@Test
			void shouldFindInAntragstellerVorname() {
				elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().antragstellerVorname("Antragsteller-Vorname").build());
				elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

				var vorgangHeaders = searchService.find(FindVorgangRequestTestFactory.createBuilder().searchBy("name").build())
						.getContent();

				assertThat(vorgangHeaders).hasSize(1);
			}
		}

		@Nested
		class filteredByOrganisationseinheitenId {

			@BeforeEach
			void init() {
				searchService.addVorgang(VorgangTestFactory.create());
				searchService.addVorgang(
						VorgangTestFactory.createBuilder().id(null).aktenzeichen("test").assignedTo(UUID.randomUUID().toString()).build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).aktenzeichen("vors").build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).aktenzeichen("vOrsi").build());
			}

			@Test
			void shouldFindByAktenzeichen() {
				var res = searchService.find(FindVorgangRequestTestFactory.createBuilder()
						.searchBy("vors")
						.filterBy(FilterCriteriaTestFactory.createBuilder().clearStatus().filterByAssignedTo(false).build())
						.build());

				assertThat(res).hasSize(2);
			}

			@Test
			void shouldFindByAntragstellerName() {
				var res = searchService
						.find(FindVorgangRequestTestFactory.createBuilder()
								.searchBy(IndexedVorgangTestFactory.ANTRAGSTELLER_NAME)
								.filterBy(FilterCriteriaTestFactory.createBuilder().clearStatus().filterByAssignedTo(false).build())
								.build());

				assertThat(res).hasSize(4);
			}

			@Test
			void shouldFindByAntragstellerNameAndAktenzeichen() {
				var res = searchService
						.find(FindVorgangRequestTestFactory.createBuilder()
								.searchBy(IndexedVorgangTestFactory.ANTRAGSTELLER_NAME + " tes")
								.filterBy(FilterCriteriaTestFactory.createBuilder().clearStatus().filterByAssignedTo(false).build())
								.build());

				assertThat(res).hasSize(1);
			}

			@Test
			void shouldFindByAntragstellerNameAndereOrganisationEinheitId() {
				searchService
						.addVorgang(
								VorgangTestFactory.createWithOrganisationEinheitId("12345678").toBuilder().id(UUID.randomUUID().toString()).build());

				var res = searchService.find(FindVorgangRequestTestFactory.createBuilder()
						.searchBy(IndexedVorgangTestFactory.ANTRAGSTELLER_NAME).filterBy(
								FilterCriteriaTestFactory.createBuilder()
										.clearOrganisationseinheitIds()
										.organisationseinheitIds(List.of("12345678"))
										.clearStatus().build())
						.build());

				assertThat(res).hasSize(1);
			}

			@Test
			void shouldDoPageination() {
				var res = searchService
						.find(FindVorgangRequestTestFactory.createBuilder()
								.searchBy(IndexedVorgangTestFactory.ANTRAGSTELLER_NAME)
								.limit(2)
								.filterBy(FilterCriteriaTestFactory.createBuilder().clearStatus().build())
								.build());

				assertThat(res).hasSize(2);
				assertThat(res.getTotalPages()).isEqualTo(2);
			}
		}

		@DisplayName("filter by status only")
		@Nested
		class TestFilterByStatus {

			@BeforeEach
			void init() {
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).status(Status.ANGENOMMEN).build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).build());
				searchService.addVorgang(VorgangTestFactory.createWithOrganisationEinheitId("12345678").toBuilder().id(null).build());
			}

			@Test
			void shouldFindByAntragstellerName() {
				var request = createRequest(IndexedVorgangTestFactory.ANTRAGSTELLER_NAME,
						FilterCriteriaTestFactory.createBuilder().filterByOrganisationseinheitenId(false).build());

				var res = searchService.find(request);

				assertThat(res).hasSize(2);
			}
		}

		@DisplayName("filter by organisationsEinheitId and status")
		@Nested
		class TestFilterByOrganisationsEinheitIdAndStatus {

			@BeforeEach
			void init() {
				searchService.addVorgang(VorgangTestFactory.create());
				searchService.addVorgang(
						VorgangTestFactory.createBuilder().id(null).aktenzeichen("test").assignedTo(UUID.randomUUID().toString()).build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).aktenzeichen("vors").build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).aktenzeichen("vOrsi").build());
				searchService
						.addVorgang(
								VorgangTestFactory.createWithOrganisationEinheitId("12345678").toBuilder().id(UUID.randomUUID().toString()).build());
				searchService
						.addVorgang(VorgangTestFactory.createBuilder().status(Status.ANGENOMMEN).id(UUID.randomUUID().toString()).build());
			}

			@Test
			void shouldFindByAktenzeichen() {
				var request = createRequest("test", FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false).build());

				var res = searchService.find(request);

				assertThat(res).hasSize(1);
			}

			@Test
			void shouldFindByAntragstellerName() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME,
						FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false).build());

				var res = searchService.find(request);

				assertThat(res).hasSize(4);
			}
		}

		private FindVorgangRequest createRequest(String searchBy, FilterCriteria filterCriteria) {
			return FindVorgangRequestTestFactory.createBuilder().searchBy(searchBy).filterBy(filterCriteria).build();
		}
	}

	@DisplayName("Filter")
	@Nested
	class TestFilter {

		@BeforeEach
		void init() {
			elasticsearchOperations.indexOps(IndexedVorgang.class).delete();
			elasticsearchOperations.indexOps(IndexedVorgang.class).create();
		}

		@DisplayName("by organisationsEinheitId")
		@Nested
		class TestByOrganisationsEinheitId {

			private static final String NAME = "shouldBeFound";

			@BeforeEach
			void init() {
				searchService.addVorgang(VorgangTestFactory.createWithOrganisationEinheitId("132").toBuilder().id(null).build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).name(NAME).assignedTo(null).build());
			}

			@Test
			void shouldReturnMatchingId() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME,
						FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false).build());

				var result = searchService.find(request);
				assertThat(result).hasSize(1);
				assertThat(result.getContent().get(0).getName()).isEqualTo(NAME);
			}

			@Test
			void shouldReturnUnfiltered() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME, FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false)
						.filterByOrganisationseinheitenId(false).organisationseinheitIds(List.of("123"))
						.build());

				var result = searchService.find(request);
				assertThat(result).hasSize(2);
			}
		}

		@DisplayName("by assignedTo")
		@Nested
		class TestByAssignedTo {

			private static final String NAME = "shouldBeFound";

			@BeforeEach
			void init() {
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).build());
				searchService.addVorgang(VorgangTestFactory.createBuilder().id(null).name(NAME).assignedTo(null).build());
			}

			@Test
			void shouldReturnUnassigned() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME,
						FilterCriteriaTestFactory.createBuilder().assignedTo(StringUtils.EMPTY).build());

				var result = searchService.find(request);

				assertThat(result).hasSize(1);
				assertThat(result.getContent().get(0).getName()).isEqualTo(NAME);
			}

			@Test
			void shouldReturnMatchingAssigned() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME, FilterCriteriaTestFactory.createBuilder().build());

				var result = searchService.find(request);

				assertThat(result).hasSize(1);
				assertThat(result.getContent().get(0).getName()).isEqualTo(VorgangTestFactory.NAME);
			}

			@Test
			void shouldReturnUnfiltered() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME,
						FilterCriteriaTestFactory.createBuilder().filterByAssignedTo(false).build());

				var result = searchService.find(request);

				assertThat(result).hasSize(2);
			}

			@Test
			void shouldFilterByAssignedToOnly() {
				var request = createRequest(AntragstellerTestFactory.NACHNAME,
						FilterCriteriaTestFactory.createBuilder().filterByOrganisationseinheitenId(false).build());

				var result = searchService.find(request);

				assertThat(result).hasSize(1);
			}
		}

		private FindVorgangRequest createRequest(String searchBy, FilterCriteria filterCriteria) {
			return FindVorgangRequestTestFactory.createBuilder().searchBy(searchBy).filterBy(filterCriteria).build();
		}
	}

	@Nested
	class TestAktenzeichen {

		@Test
		void shouldSetAktenzeichenOnly() {
			var newAktenzeichen = "new-aktenzeichen";
			var vorgang = elasticsearchOperations.save(IndexedVorgangTestFactory.createBuilder().vorgangId(null).build());
			elasticsearchOperations.indexOps(IndexedVorgang.class).refresh();

			searchService.setAktenzeichen(vorgang.getVorgangId(), newAktenzeichen);

			var indexedVorgang = elasticsearchOperations.get(vorgang.getVorgangId(), IndexedVorgang.class);
			assertThat(indexedVorgang).usingRecursiveComparison().ignoringFields("vorgangId").isEqualTo(IndexedVorgangTestFactory.createBuilder().aktenzeichen(newAktenzeichen).build());
		}

	}
}
