/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.List;

import org.apache.commons.lang3.StringUtils;
import org.bson.Document;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;

import static org.assertj.core.api.Assertions.*;

import de.ozgcloud.vorgang.callcontext.CallContextTestFactory;
import de.ozgcloud.vorgang.callcontext.UserTestFactory;
import de.ozgcloud.vorgang.vorgang.IncomingFileTestFactory;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;

class BinaryFileRepositoryTest {

	@Spy
	@InjectMocks
	private BinaryFileRepository grisFsService = new BinaryFileRepository();
	@Mock
	private GridFsTemplate gridFsTemplate;

	@Nested
	class TestMetaDataCreation {

		@Test
		void shouldCreateMetaData() {
			var doc = callCreateMetadata();

			assertThat(doc).isNotNull();
		}

		@Test
		void shouldHaveVorgangId() {
			var doc = callCreateMetadata();

			assertThat(doc).containsEntry(BinaryFileRepository.VORGANG_ID, UploadedFilesReferenceTestFactory.VORGANG_ID);
		}

		@Test
		void shouldHaveClient() {
			var doc = callCreateMetadata();

			assertThat(doc).containsEntry(BinaryFileRepository.CLIENT, CallContextTestFactory.CLIENT);
		}

		@Test
		void shouldHaveContentType() {
			var doc = callCreateMetadata();

			assertThat(doc).containsEntry(BinaryFileRepository.CONTENT_TYPE, OzgFileTestFactory.CONTENT_TYPE);
		}

		@Test
		void shouldHaveCreatedBY() {
			var doc = callCreateMetadata();

			assertThat(doc).containsEntry(BinaryFileRepository.CREATED_BY, UserTestFactory.ID);
		}

		@Test
		void shouldHaveFileName() {
			var doc = callCreateMetadata();

			assertThat(doc).containsEntry(BinaryFileRepository.FILE_NAME, OzgFileTestFactory.NAME);
		}

		@Test
		void shouldHaveFieldName() {
			var doc = callCreateMetadata();

			assertThat(doc).containsEntry(BinaryFileRepository.FIELD_NAME, UploadedFilesReferenceTestFactory.NAME);
		}

		private Document callCreateMetadata() {
			return grisFsService.createMetaData(UploadedFilesReferenceTestFactory.create(), OzgFileTestFactory.create(), UserTestFactory.ID);
		}
	}

	@Nested
	class TestFilePath {
		@Test
		void shouldCreatePath() {
			String path = grisFsService.createFilePath(
					UploadedFilesReferenceTestFactory.create(),
					OzgFileTestFactory.create());

			assertThat(path).isEqualTo(
					UploadedFilesReferenceTestFactory.VORGANG_ID + "/" +
							CallContextTestFactory.CLIENT + "/" +
							UploadedFilesReferenceTestFactory.NAME + "/" +
							OzgFileTestFactory.NAME);
		}
	}

	@Nested
	class TestFindAllMetaData {

		@Test
		void shouldCallGridFSTemplate() {
			when(gridFsTemplate.findOne(any())).thenReturn(GridFsTestFactory.create());

			grisFsService.getAll(Collections.singleton(IncomingFileTestFactory.ID.toString())).toList();

			verify(gridFsTemplate).findOne(any());
		}

		@Test
		void shouldCallGridFsTemplateFind() {
			when(gridFsTemplate.findOne(any())).thenReturn(GridFsTestFactory.create());

			var files = grisFsService.getAll(Collections.singleton(IncomingFileTestFactory.ID.toString())).toList();

			assertThat(files).hasSize(1);
		}

		@Test
		@DisplayName("should not map to OZG File when FileId is empty")
		void shouldNotMapWhenFileIdEmpty() {
			grisFsService.getAll(List.of(StringUtils.EMPTY)).toList();

			verify(grisFsService, never()).mapToOzgFile(any());
		}

		@Test
		@DisplayName("should not map to OZG File when FileId is null")
		void shouldNotMapWhenFileIdNull() {
			grisFsService.getAll(Collections.singletonList(null)).toList();

			verify(grisFsService, never()).mapToOzgFile(any());
		}

		@Nested
		class TestMapToOzgFile {

			@Test
			void shouldContainFileId() {
				var ozgFile = callRepositoryMap();

				assertThat(ozgFile.getId()).hasToString(GridFsTestFactory.ID.toHexString());
			}

			@Test
			void shouldContainVorgangId() {
				var ozgFile = callRepositoryMap();

				assertThat(ozgFile.getVorgangId()).isEqualTo(VorgangTestFactory.ID);
			}

			@Test
			void shouldContainName() {
				var ozgFile = callRepositoryMap();

				assertThat(ozgFile.getName()).hasToString(OzgFileTestFactory.NAME);
			}

			@Test
			void shouldContainSize() {
				var ozgFile = callRepositoryMap();

				assertThat(ozgFile.getSize()).isEqualTo(OzgFileTestFactory.SIZE);
			}

			@Test
			void shouldContainContentType() {
				var ozgFile = callRepositoryMap();

				assertThat(ozgFile.getContentType()).hasToString(OzgFileTestFactory.CONTENT_TYPE);
			}

			private OzgFile callRepositoryMap() {
				return grisFsService.mapToOzgFile(GridFsTestFactory.create());
			}
		}
	}
}
